@@ -50,7 +50,9 @@ __all__ = [
     'Constraint',
 ]
 
-import jax.numpy as jnp
+import numpy as np
+
+import jax.numpy
 
 
 class Constraint(object):
@@ -60,6 +62,7 @@ class Constraint(object):
     A constraint object represents a region over which a variable is valid,
     e.g. within which a variable can be optimized.
     """
+    event_dim = 0
 
     def __call__(self, x):
         raise NotImplementedError
@@ -71,14 +74,26 @@ class Constraint(object):
         """
         return self(value)
 
+    def feasible_like(self, prototype):
+        """
+        Get a feasible value which has the same shape as dtype as `prototype`.
+        """
+        raise NotImplementedError
+
 
 class _Boolean(Constraint):
     def __call__(self, x):
         return (x == 0) | (x == 1)
 
+    def feasible_like(self, prototype):
+        return jax.numpy.zeros_like(prototype)
+
 
 class _CorrCholesky(Constraint):
+    event_dim = 2
+
     def __call__(self, x):
+        jnp = np if isinstance(x, (np.ndarray, np.generic)) else jax.numpy
         tril = jnp.tril(x)
         lower_triangular = jnp.all(jnp.reshape(tril == x, x.shape[:-2] + (-1,)), axis=-1)
         positive_diagonal = jnp.all(jnp.diagonal(x, axis1=-2, axis2=-1) > 0, axis=-1)
@@ -86,9 +101,15 @@ class _CorrCholesky(Constraint):
         unit_norm_row = jnp.all((x_norm <= 1) & (x_norm > 1 - 1e-6), axis=-1)
         return lower_triangular & positive_diagonal & unit_norm_row
 
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to(jax.numpy.eye(prototype.shape[-1]), prototype.shape)
+
 
 class _CorrMatrix(Constraint):
+    event_dim = 2
+
     def __call__(self, x):
+        jnp = np if isinstance(x, (np.ndarray, np.generic)) else jax.numpy
         # check for symmetric
         symmetric = jnp.all(jnp.all(x == jnp.swapaxes(x, -2, -1), axis=-1), axis=-1)
         # check for the smallest eigenvalue is positive
@@ -97,6 +118,9 @@ class _CorrMatrix(Constraint):
         unit_variance = jnp.all(jnp.abs(jnp.diagonal(x, axis1=-2, axis2=-1) - 1) < 1e-6, axis=-1)
         return symmetric & positive & unit_variance
 
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to(jax.numpy.eye(prototype.shape[-1]), prototype.shape)
+
 
 class _Dependent(Constraint):
     def __call__(self, x):
@@ -114,6 +138,46 @@ class _GreaterThan(Constraint):
     def __call__(self, x):
         return x > self.lower_bound
 
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to(self.lower_bound + 1, jax.numpy.shape(prototype))
+
+
+class _IndependentConstraint(Constraint):
+    """
+    Wraps a constraint by aggregating over ``reinterpreted_batch_ndims``-many
+    dims in :meth:`check`, so that an event is valid only if all its
+    independent entries are valid.
+    """
+    def __init__(self, base_constraint, reinterpreted_batch_ndims):
+        assert isinstance(base_constraint, Constraint)
+        assert isinstance(reinterpreted_batch_ndims, int)
+        assert reinterpreted_batch_ndims >= 0
+        if isinstance(base_constraint, _IndependentConstraint):
+            reinterpreted_batch_ndims = reinterpreted_batch_ndims + base_constraint.reinterpreted_batch_ndims
+            base_constraint = base_constraint.base_constraint
+        self.base_constraint = base_constraint
+        self.reinterpreted_batch_ndims = reinterpreted_batch_ndims
+        super().__init__()
+
+    @property
+    def event_dim(self):
+        return self.base_constraint.event_dim + self.reinterpreted_batch_ndims
+
+    def __call__(self, value):
+        result = self.base_constraint(value)
+        if self.reinterpreted_batch_ndims == 0:
+            return result
+        elif jax.numpy.ndim(result) < self.reinterpreted_batch_ndims:
+            expected = self.event_dim
+            raise ValueError(f"Expected value.dim() >= {expected} but got {jax.numpy.ndim(value)}")
+        result = result.reshape(
+            jax.numpy.shape(result)[:jax.numpy.ndim(result) - self.reinterpreted_batch_ndims] + (-1,))
+        result = result.all(-1)
+        return result
+
+    def feasible_like(self, prototype):
+        return self.base_constraint.feasible_like(prototype)
+
 
 class _LessThan(Constraint):
     def __init__(self, upper_bound):
@@ -122,6 +186,9 @@ class _LessThan(Constraint):
     def __call__(self, x):
         return x < self.upper_bound
 
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to(self.upper_bound - 1, jax.numpy.shape(prototype))
+
 
 class _IntegerInterval(Constraint):
     def __init__(self, lower_bound, upper_bound):
@@ -129,7 +196,10 @@ class _IntegerInterval(Constraint):
         self.upper_bound = upper_bound
 
     def __call__(self, x):
-        return (x >= self.lower_bound) & (x <= self.upper_bound) & (x == jnp.floor(x))
+        return (x >= self.lower_bound) & (x <= self.upper_bound) & (x % 1 == 0)
+
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to(self.lower_bound, jax.numpy.shape(prototype))
 
 
 class _IntegerGreaterThan(Constraint):
@@ -139,6 +209,9 @@ class _IntegerGreaterThan(Constraint):
     def __call__(self, x):
         return (x % 1 == 0) & (x >= self.lower_bound)
 
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to(self.lower_bound, jax.numpy.shape(prototype))
+
 
 class _Interval(Constraint):
     def __init__(self, lower_bound, upper_bound):
@@ -146,53 +219,84 @@ class _Interval(Constraint):
         self.upper_bound = upper_bound
 
     def __call__(self, x):
-        return (x > self.lower_bound) & (x < self.upper_bound)
+        return (x >= self.lower_bound) & (x <= self.upper_bound)
+
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to((self.lower_bound + self.upper_bound) / 2, jax.numpy.shape(prototype))
 
 
 class _LowerCholesky(Constraint):
+    event_dim = 2
+
     def __call__(self, x):
+        jnp = np if isinstance(x, (np.ndarray, np.generic)) else jax.numpy
         tril = jnp.tril(x)
         lower_triangular = jnp.all(jnp.reshape(tril == x, x.shape[:-2] + (-1,)), axis=-1)
         positive_diagonal = jnp.all(jnp.diagonal(x, axis1=-2, axis2=-1) > 0, axis=-1)
         return lower_triangular & positive_diagonal
 
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to(jax.numpy.eye(prototype.shape[-1]), prototype.shape)
+
 
 class _Multinomial(Constraint):
+    event_dim = 1
+
     def __init__(self, upper_bound):
         self.upper_bound = upper_bound
 
     def __call__(self, x):
-        return jnp.all(x >= 0, axis=-1) & (jnp.sum(x, -1) == self.upper_bound)
+        return (x >= 0).all(axis=-1) & (x.sum(axis=-1) == self.upper_bound)
+
+    def feasible_like(self, prototype):
+        pad_width = ((0, 0),) * jax.numpy.ndim(self.upper_bound) + ((0, prototype.shape[-1] - 1),)
+        value = jax.numpy.pad(jax.numpy.expand_dims(self.upper_bound, -1), pad_width)
+        return jax.numpy.broadcast_to(value, prototype.shape)
 
 
 class _OrderedVector(Constraint):
+    event_dim = 1
+
     def __call__(self, x):
-        return jnp.all(x[..., 1:] > x[..., :-1], axis=-1)
+        return (x[..., 1:] > x[..., :-1]).all(axis=-1)
+
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to(jax.numpy.arange(float(prototype.shape[-1])), prototype.shape)
 
 
 class _PositiveDefinite(Constraint):
+    event_dim = 2
+
     def __call__(self, x):
+        jnp = np if isinstance(x, (np.ndarray, np.generic)) else jax.numpy
         # check for symmetric
         symmetric = jnp.all(jnp.all(x == jnp.swapaxes(x, -2, -1), axis=-1), axis=-1)
         # check for the smallest eigenvalue is positive
         positive = jnp.linalg.eigh(x)[0][..., 0] > 0
         return symmetric & positive
 
+    def feasible_like(self, prototype):
+        return jax.numpy.broadcast_to(jax.numpy.eye(prototype.shape[-1]), prototype.shape)
+
 
 class _Real(Constraint):
     def __call__(self, x):
-        return jnp.isfinite(x)
+        # XXX: consider to relax this condition to [-inf, inf] interval
+        return (x == x) & (x != float('inf')) & (x != float('-inf'))
 
-
-class _RealVector(Constraint):
-    def __call__(self, x):
-        return jnp.all(jnp.isfinite(x), axis=-1)
+    def feasible_like(self, prototype):
+        return jax.numpy.zeros_like(prototype)
 
 
 class _Simplex(Constraint):
+    event_dim = 1
+
     def __call__(self, x):
-        x_sum = jnp.sum(x, axis=-1)
-        return jnp.all(x > 0, axis=-1) & (x_sum <= 1) & (x_sum > 1 - 1e-6)
+        x_sum = x.sum(axis=-1)
+        return (x >= 0).all(axis=-1) & (x_sum < 1 + 1e-6) & (x_sum > 1 - 1e-6)
+
+    def feasible_like(self, prototype):
+        return jax.numpy.full_like(prototype, 1 / prototype.shape[-1])
 
 
 # TODO: Make types consistent
@@ -203,6 +307,7 @@ corr_matrix = _CorrMatrix()
 dependent = _Dependent()
 greater_than = _GreaterThan
 less_than = _LessThan
+independent = _IndependentConstraint
 integer_interval = _IntegerInterval
 integer_greater_than = _IntegerGreaterThan
 interval = _Interval
@@ -214,6 +319,6 @@ positive = _GreaterThan(0.)
 positive_definite = _PositiveDefinite()
 positive_integer = _IntegerGreaterThan(1)
 real = _Real()
-real_vector = _RealVector()
+real_vector = independent(real, 1)
 simplex = _Simplex()
 unit_interval = _Interval(0., 1.)
