@@ -38,6 +38,7 @@ from numpyro.distributions.distribution import Distribution, TransformedDistribu
 from numpyro.distributions.transforms import AffineTransform, ExpTransform, InvCholeskyTransform, PowerTransform
 from numpyro.distributions.util import (
     cholesky_of_inverse,
+    is_prng_key,
     lazy_property,
     matrix_to_tril_vec,
     promote_shapes,
@@ -51,17 +52,20 @@ EULER_MASCHERONI = 0.5772156649015328606065120900824024310421
 
 class Beta(Distribution):
     arg_constraints = {'concentration1': constraints.positive, 'concentration0': constraints.positive}
+    reparametrized_params = ['concentration1', 'concentration0']
     support = constraints.unit_interval
 
     def __init__(self, concentration1, concentration0, validate_args=None):
+        self.concentration1, self.concentration0 = promote_shapes(concentration1, concentration0)
         batch_shape = lax.broadcast_shapes(jnp.shape(concentration1), jnp.shape(concentration0))
-        self.concentration1 = jnp.broadcast_to(concentration1, batch_shape)
-        self.concentration0 = jnp.broadcast_to(concentration0, batch_shape)
-        self._dirichlet = Dirichlet(jnp.stack([self.concentration1, self.concentration0],
+        concentration1 = jnp.broadcast_to(concentration1, batch_shape)
+        concentration0 = jnp.broadcast_to(concentration0, batch_shape)
+        self._dirichlet = Dirichlet(jnp.stack([concentration1, concentration0],
                                               axis=-1))
         super(Beta, self).__init__(batch_shape=batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return self._dirichlet.sample(key, sample_shape)[..., 0]
 
     @validate_sample
@@ -89,6 +93,7 @@ class Cauchy(Distribution):
         super(Cauchy, self).__init__(batch_shape=batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         eps = random.cauchy(key, shape=sample_shape + self.batch_shape)
         return self.loc + eps * self.scale
 
@@ -107,6 +112,7 @@ class Cauchy(Distribution):
 
 class Dirichlet(Distribution):
     arg_constraints = {'concentration': constraints.positive}
+    reparametrized_params = ['concentration']
     support = constraints.simplex
 
     def __init__(self, concentration, validate_args=None):
@@ -119,6 +125,7 @@ class Dirichlet(Distribution):
                                         validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         shape = sample_shape + self.batch_shape + self.event_shape
         gamma_samples = random.gamma(key, self.concentration, shape=shape)
         samples = gamma_samples / jnp.sum(gamma_samples, axis=-1, keepdims=True)
@@ -150,6 +157,7 @@ class Exponential(Distribution):
         super(Exponential, self).__init__(batch_shape=jnp.shape(rate), validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return random.exponential(key, shape=sample_shape + self.batch_shape) / self.rate
 
     @validate_sample
@@ -169,7 +177,7 @@ class Gamma(Distribution):
     arg_constraints = {'concentration': constraints.positive,
                        'rate': constraints.positive}
     support = constraints.positive
-    reparametrized_params = ['rate']
+    reparametrized_params = ['concentration', 'rate']
 
     def __init__(self, concentration, rate=1., validate_args=None):
         self.concentration, self.rate = promote_shapes(concentration, rate)
@@ -178,6 +186,7 @@ class Gamma(Distribution):
                                     validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         shape = sample_shape + self.batch_shape + self.event_shape
         return random.gamma(key, self.concentration, shape=shape) / self.rate
 
@@ -198,6 +207,7 @@ class Gamma(Distribution):
 
 class Chi2(Gamma):
     arg_constraints = {'df': constraints.positive}
+    reparametrized_params = ['df']
 
     def __init__(self, df, validate_args=None):
         self.df = df
@@ -205,18 +215,20 @@ class Chi2(Gamma):
 
 
 class GaussianRandomWalk(Distribution):
-    arg_constraints = {'scale': constraints.positive, 'num_steps': constraints.positive_integer}
+    arg_constraints = {'scale': constraints.positive}
     support = constraints.real_vector
     reparametrized_params = ['scale']
 
     def __init__(self, scale=1., num_steps=1, validate_args=None):
-        assert jnp.shape(num_steps) == ()
+        assert isinstance(num_steps, int) and num_steps > 0, \
+            "`num_steps` argument should be an positive integer."
         self.scale = scale
         self.num_steps = num_steps
         batch_shape, event_shape = jnp.shape(scale), (num_steps,)
         super(GaussianRandomWalk, self).__init__(batch_shape, event_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         shape = sample_shape + self.batch_shape + self.event_shape
         walks = random.normal(key, shape=shape)
         return jnp.cumsum(walks, axis=-1) * jnp.expand_dims(self.scale, axis=-1)
@@ -256,6 +268,7 @@ class HalfCauchy(Distribution):
         super(HalfCauchy, self).__init__(batch_shape=jnp.shape(scale), validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return jnp.abs(self._cauchy.sample(key, sample_shape))
 
     @validate_sample
@@ -282,6 +295,7 @@ class HalfNormal(Distribution):
         super(HalfNormal, self).__init__(batch_shape=jnp.shape(scale), validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return jnp.abs(self._normal.sample(key, sample_shape))
 
     @validate_sample
@@ -298,14 +312,16 @@ class HalfNormal(Distribution):
 
 
 class InverseGamma(TransformedDistribution):
+    """
+    .. note:: We keep the same notation `rate` as in Pyro but
+        it plays the role of scale parameter of InverseGamma in literatures
+        (e.g. wikipedia: https://en.wikipedia.org/wiki/Inverse-gamma_distribution)
+    """
     arg_constraints = {'concentration': constraints.positive, 'rate': constraints.positive}
+    reparametrized_params = ["concentration", "rate"]
     support = constraints.positive
-    reparametrized_params = ['rate']
 
     def __init__(self, concentration, rate=1., validate_args=None):
-        # NB: we keep the same notation `rate` as in Pyro and tensorflow but
-        # it plays the role of scale parameter of InverseGamma in literatures
-        # (e.g. wikipedia: https://en.wikipedia.org/wiki/Inverse-gamma_distribution)
         base_dist = Gamma(concentration, rate)
         self.concentration = base_dist.concentration
         self.rate = base_dist.rate
@@ -341,6 +357,7 @@ class Gumbel(Distribution):
                                      validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         standard_gumbel_sample = random.gumbel(key, shape=sample_shape + self.batch_shape + self.event_shape)
         return self.loc + self.scale * standard_gumbel_sample
 
@@ -371,6 +388,7 @@ class Laplace(Distribution):
         super(Laplace, self).__init__(batch_shape=batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         eps = random.laplace(key, shape=sample_shape + self.batch_shape + self.event_shape)
         return self.loc + eps * self.scale
 
@@ -415,6 +433,7 @@ class LKJ(TransformedDistribution):
     Daniel Lewandowski, Dorota Kurowicka, Harry Joe
     """
     arg_constraints = {'concentration': constraints.positive}
+    reparametrized_params = ["concentration"]
     support = constraints.corr_matrix
 
     def __init__(self, dimension, concentration=1., sample_method='onion', validate_args=None):
@@ -466,6 +485,7 @@ class LKJCholesky(Distribution):
     Daniel Lewandowski, Dorota Kurowicka, Harry Joe
     """
     arg_constraints = {'concentration': constraints.positive}
+    reparametrized_params = ['concentration']
     support = constraints.corr_cholesky
 
     def __init__(self, dimension, concentration=1., sample_method='onion', validate_args=None):
@@ -551,6 +571,7 @@ class LKJCholesky(Distribution):
         return cholesky
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         if self.sample_method == "onion":
             return self._onion(key, sample_shape)
         else:
@@ -685,8 +706,8 @@ class MultivariateNormal(Distribution):
 
     def __init__(self, loc=0., covariance_matrix=None, precision_matrix=None, scale_tril=None,
                  validate_args=None):
-        if jnp.isscalar(loc):
-            loc = jnp.expand_dims(loc, axis=-1)
+        if jnp.ndim(loc) == 0:
+            loc, = promote_shapes(loc, shape=(1,))
         # temporary append a new axis to loc
         loc = loc[..., jnp.newaxis]
         if covariance_matrix is not None:
@@ -702,12 +723,13 @@ class MultivariateNormal(Distribution):
                              ' must be specified.')
         batch_shape = lax.broadcast_shapes(jnp.shape(loc)[:-2], jnp.shape(self.scale_tril)[:-2])
         event_shape = jnp.shape(self.scale_tril)[-1:]
-        self.loc = jnp.broadcast_to(jnp.squeeze(loc, axis=-1), batch_shape + event_shape)
+        self.loc = loc[..., 0]
         super(MultivariateNormal, self).__init__(batch_shape=batch_shape,
                                                  event_shape=event_shape,
                                                  validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         eps = random.normal(key, shape=sample_shape + self.batch_shape + self.event_shape)
         return self.loc + jnp.squeeze(jnp.matmul(self.scale_tril, eps[..., jnp.newaxis]), axis=-1)
 
@@ -729,7 +751,7 @@ class MultivariateNormal(Distribution):
 
     @property
     def mean(self):
-        return self.loc
+        return jnp.broadcast_to(self.loc, self.shape())
 
     @property
     def variance(self):
@@ -795,10 +817,11 @@ def _batch_lowrank_mahalanobis(W, D, x, capacitance_tril):
 class LowRankMultivariateNormal(Distribution):
     arg_constraints = {
         "loc": constraints.real_vector,
-        "cov_factor": constraints.real,
-        "cov_diag": constraints.positive
-        }
+        "cov_factor": constraints.independent(constraints.real, 2),
+        "cov_diag": constraints.independent(constraints.positive, 1)
+    }
     support = constraints.real_vector
+    reparametrized_params = ['loc', 'cov_factor', 'cov_diag']
 
     def __init__(self, loc, cov_factor, cov_diag, validate_args=None):
         if jnp.ndim(loc) < 1:
@@ -815,7 +838,7 @@ class LowRankMultivariateNormal(Distribution):
 
         loc, cov_factor, cov_diag = promote_shapes(loc[..., jnp.newaxis], cov_factor, cov_diag[..., jnp.newaxis])
         batch_shape = lax.broadcast_shapes(jnp.shape(loc), jnp.shape(cov_factor), jnp.shape(cov_diag))[:-2]
-        self.loc = jnp.broadcast_to(loc[..., 0], batch_shape + event_shape)
+        self.loc = loc[..., 0]
         self.cov_factor = cov_factor
         cov_diag = cov_diag[..., 0]
         self.cov_diag = cov_diag
@@ -870,6 +893,7 @@ class LowRankMultivariateNormal(Distribution):
         return diag_embed - jnp.matmul(jnp.swapaxes(A, -1, -2), A)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         key_W, key_D = random.split(key)
         batch_shape = sample_shape + self.batch_shape
         W_shape = batch_shape + self.cov_factor.shape[-1:]
@@ -910,6 +934,7 @@ class Normal(Distribution):
         super(Normal, self).__init__(batch_shape=batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         eps = random.normal(key, shape=sample_shape + self.batch_shape + self.event_shape)
         return self.loc + eps * self.scale
 
@@ -933,24 +958,26 @@ class Normal(Distribution):
 
 class Pareto(TransformedDistribution):
     arg_constraints = {'scale': constraints.positive, 'alpha': constraints.positive}
+    reparametrized_params = ["scale", "alpha"]
 
     def __init__(self, scale, alpha, validate_args=None):
+        self.scale, self.alpha = promote_shapes(scale, alpha)
         batch_shape = lax.broadcast_shapes(jnp.shape(scale), jnp.shape(alpha))
-        self.scale, self.alpha = jnp.broadcast_to(scale, batch_shape), jnp.broadcast_to(alpha, batch_shape)
-        base_dist = Exponential(self.alpha)
-        transforms = [ExpTransform(), AffineTransform(loc=0, scale=self.scale)]
+        scale, alpha = jnp.broadcast_to(scale, batch_shape), jnp.broadcast_to(alpha, batch_shape)
+        base_dist = Exponential(alpha)
+        transforms = [ExpTransform(), AffineTransform(loc=0, scale=scale)]
         super(Pareto, self).__init__(base_dist, transforms, validate_args=validate_args)
 
     @property
     def mean(self):
         # mean is inf for alpha <= 1
-        a = lax.div(self.alpha * self.scale, (self.alpha - 1))
+        a = jnp.divide(self.alpha * self.scale, (self.alpha - 1))
         return jnp.where(self.alpha <= 1, jnp.inf, a)
 
     @property
     def variance(self):
         # var is inf for alpha <= 2
-        a = lax.div((self.scale ** 2) * self.alpha, (self.alpha - 1) ** 2 * (self.alpha - 2))
+        a = jnp.divide((self.scale ** 2) * self.alpha, (self.alpha - 1) ** 2 * (self.alpha - 2))
         return jnp.where(self.alpha <= 2, jnp.inf, a)
 
     # override the default behaviour to save computations
@@ -965,16 +992,17 @@ class Pareto(TransformedDistribution):
 class StudentT(Distribution):
     arg_constraints = {'df': constraints.positive, 'loc': constraints.real, 'scale': constraints.positive}
     support = constraints.real
-    reparametrized_params = ['loc', 'scale']
+    reparametrized_params = ['df', 'loc', 'scale']
 
     def __init__(self, df, loc=0., scale=1., validate_args=None):
         batch_shape = lax.broadcast_shapes(jnp.shape(df), jnp.shape(loc), jnp.shape(scale))
-        self.df = jnp.broadcast_to(df, batch_shape)
-        self.loc, self.scale = promote_shapes(loc, scale, shape=batch_shape)
-        self._chi2 = Chi2(self.df)
+        self.df, self.loc, self.scale = promote_shapes(df, loc, scale, shape=batch_shape)
+        df = jnp.broadcast_to(df, batch_shape)
+        self._chi2 = Chi2(df)
         super(StudentT, self).__init__(batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         key_normal, key_chi2 = random.split(key)
         std_normal = random.normal(key_normal, shape=sample_shape + self.batch_shape)
         z = self._chi2.sample(key_chi2, sample_shape)
@@ -995,13 +1023,15 @@ class StudentT(Distribution):
 
     @property
     def variance(self):
-        var = jnp.where(self.df > 2, self.scale ** 2 * self.df / (self.df - 2.0), jnp.inf)
+        var = jnp.where(self.df > 2, jnp.divide(self.scale ** 2 * self.df, self.df - 2.0), jnp.inf)
         var = jnp.where(self.df <= 1, jnp.nan, var)
         return jnp.broadcast_to(var, self.batch_shape)
 
 
 class _BaseTruncatedCauchy(Distribution):
     # NB: this is a truncated cauchy with low=0, scale=1
+    arg_constraints = {"base_loc": constraints.real}
+    reparametrized_params = ["base_loc"]
     support = constraints.positive
 
     def __init__(self, base_loc):
@@ -1009,6 +1039,7 @@ class _BaseTruncatedCauchy(Distribution):
         super(_BaseTruncatedCauchy, self).__init__(batch_shape=jnp.shape(base_loc))
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         # We use inverse transform method:
         # z ~ inv_cdf(U), where U ~ Uniform(cdf(low), cdf(high)).
         #                         ~ Uniform(arctan(low), arctan(high)) / pi + 1/2
@@ -1028,7 +1059,7 @@ class _BaseTruncatedCauchy(Distribution):
 class TruncatedCauchy(TransformedDistribution):
     arg_constraints = {'low': constraints.real, 'loc': constraints.real,
                        'scale': constraints.positive}
-    reparametrized_params = ['low', 'loc', 'scale']
+    reparametrized_params = ["low", "loc", "scale"]
 
     def __init__(self, low=0., loc=0., scale=1., validate_args=None):
         self.low, self.loc, self.scale = promote_shapes(low, loc, scale)
@@ -1068,6 +1099,8 @@ class TruncatedCauchy(TransformedDistribution):
 
 class _BaseTruncatedNormal(Distribution):
     # NB: this is a truncated normal with low=0, scale=1
+    arg_constraints = {"base_loc": constraints.real}
+    reparametrized_params = ["base_loc"]
     support = constraints.positive
 
     def __init__(self, base_loc):
@@ -1076,6 +1109,7 @@ class _BaseTruncatedNormal(Distribution):
         super(_BaseTruncatedNormal, self).__init__(batch_shape=jnp.shape(base_loc))
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         size = sample_shape + self.batch_shape
         # We use inverse transform method:
         # z ~ icdf(U), where U ~ Uniform(0, 1).
@@ -1094,7 +1128,7 @@ class _BaseTruncatedNormal(Distribution):
 class TruncatedNormal(TransformedDistribution):
     arg_constraints = {'low': constraints.real, 'loc': constraints.real,
                        'scale': constraints.positive}
-    reparametrized_params = ['low', 'loc', 'scale']
+    reparametrized_params = ["low", "loc", "scale"]
 
     # TODO: support `high` arg
     def __init__(self, low=0., loc=0., scale=1., validate_args=None):
@@ -1141,6 +1175,7 @@ class _BaseUniform(Distribution):
         super(_BaseUniform, self).__init__(batch_shape=batch_shape)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         size = sample_shape + self.batch_shape
         return random.uniform(key, shape=size)
 
@@ -1192,7 +1227,7 @@ class Uniform(TransformedDistribution):
 class Logistic(Distribution):
     arg_constraints = {'loc': constraints.real, 'scale': constraints.positive}
     support = constraints.real
-    reparametrized_params = ['loc', 'real']
+    reparametrized_params = ['loc', 'scale']
 
     def __init__(self, loc=0., scale=1., validate_args=None):
         self.loc, self.scale = promote_shapes(loc, scale)
@@ -1200,6 +1235,7 @@ class Logistic(Distribution):
         super(Logistic, self).__init__(batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         z = random.logistic(key, shape=sample_shape + self.batch_shape + self.event_shape)
         return self.loc + z * self.scale
 
@@ -1232,6 +1268,7 @@ class TruncatedPolyaGamma(Distribution):
         super(TruncatedPolyaGamma, self).__init__(batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         denom = jnp.square(jnp.arange(0.5, self.num_gamma_variates))
         x = random.gamma(key, jnp.ones(self.batch_shape + sample_shape + (self.num_gamma_variates,)))
         x = jnp.sum(x / denom, axis=-1)
