@@ -25,11 +25,12 @@
 # ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 
+import warnings
+
 import numpy as np
 
 from jax import device_put, lax
-from jax.dtypes import canonicalize_dtype
-from jax.nn import softmax
+from jax.nn import softmax, softplus
 import jax.numpy as jnp
 import jax.random as random
 from jax.scipy.special import expit, gammaln, logsumexp, xlog1py, xlogy
@@ -42,17 +43,17 @@ from numpyro.distributions.util import (
     categorical,
     clamp_probs,
     get_dtype,
+    is_prng_key,
     lazy_property,
     multinomial,
     promote_shapes,
-    sum_rightmost,
     validate_sample
 )
 from numpyro.util import not_jax_tracer
 
 
 def _to_probs_bernoulli(logits):
-    return 1 / (1 + jnp.exp(-logits))
+    return expit(logits)
 
 
 def _to_logits_bernoulli(probs):
@@ -80,12 +81,18 @@ class BernoulliProbs(Distribution):
         super(BernoulliProbs, self).__init__(batch_shape=jnp.shape(self.probs), validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
-        return random.bernoulli(key, self.probs, shape=sample_shape + self.batch_shape)
+        assert is_prng_key(key)
+        samples = random.bernoulli(key, self.probs, shape=sample_shape + self.batch_shape)
+        return samples.astype(jnp.result_type(samples, int))
 
     @validate_sample
     def log_prob(self, value):
         return xlogy(value, self.probs) + xlog1py(1 - value, -self.probs)
 
+    @lazy_property
+    def logits(self):
+        return _to_logits_bernoulli(self.probs)
+
     @property
     def mean(self):
         return self.probs
@@ -112,7 +119,9 @@ class BernoulliLogits(Distribution):
         super(BernoulliLogits, self).__init__(batch_shape=jnp.shape(self.logits), validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
-        return random.bernoulli(key, self.probs, shape=sample_shape + self.batch_shape)
+        assert is_prng_key(key)
+        samples = random.bernoulli(key, self.probs, shape=sample_shape + self.batch_shape)
+        return samples.astype(jnp.result_type(samples, int))
 
     @validate_sample
     def log_prob(self, value):
@@ -158,6 +167,7 @@ class BinomialProbs(Distribution):
         super(BinomialProbs, self).__init__(batch_shape=batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return binomial(key, self.probs, n=self.total_count, shape=sample_shape + self.batch_shape)
 
     @validate_sample
@@ -168,6 +178,10 @@ class BinomialProbs(Distribution):
         return (log_factorial_n - log_factorial_k - log_factorial_nmk +
                 xlogy(value, self.probs) + xlog1py(self.total_count - value, -self.probs))
 
+    @lazy_property
+    def logits(self):
+        return _to_logits_bernoulli(self.probs)
+
     @property
     def mean(self):
         return jnp.broadcast_to(self.total_count * self.probs, self.batch_shape)
@@ -181,12 +195,14 @@ class BinomialProbs(Distribution):
         return constraints.integer_interval(0, self.total_count)
 
     def enumerate_support(self, expand=True):
-        total_count = jnp.amax(self.total_count)
-        if not_jax_tracer(total_count):
+        if not_jax_tracer(self.total_count):
+            total_count = np.amax(self.total_count)
             # NB: the error can't be raised if inhomogeneous issue happens when tracing
-            if jnp.amin(self.total_count) != total_count:
+            if np.amin(self.total_count) != total_count:
                 raise NotImplementedError("Inhomogeneous total count not supported"
                                           " by `enumerate_support`.")
+        else:
+            total_count = jnp.amax(self.total_count)
         values = jnp.arange(total_count + 1).reshape((-1,) + (1,) * len(self.batch_shape))
         if expand:
             values = jnp.broadcast_to(values, values.shape[:1] + self.batch_shape)
@@ -198,6 +214,7 @@ class BinomialLogits(Distribution):
                        'total_count': constraints.nonnegative_integer}
     has_enumerate_support = True
     is_discrete = True
+    enumerate_support = BinomialProbs.enumerate_support
 
     def __init__(self, logits, total_count=1, validate_args=None):
         self.logits, self.total_count = promote_shapes(logits, total_count)
@@ -205,6 +222,7 @@ class BinomialLogits(Distribution):
         super(BinomialLogits, self).__init__(batch_shape=batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return binomial(key, self.probs, n=self.total_count, shape=sample_shape + self.batch_shape)
 
     @validate_sample
@@ -233,18 +251,6 @@ class BinomialLogits(Distribution):
     def support(self):
         return constraints.integer_interval(0, self.total_count)
 
-    def enumerate_support(self, expand=True):
-        total_count = jnp.amax(self.total_count)
-        if not_jax_tracer(total_count):
-            # NB: the error can't be raised if inhomogeneous issue happens when tracing
-            if jnp.amin(self.total_count) != total_count:
-                raise NotImplementedError("Inhomogeneous total count not supported"
-                                          " by `enumerate_support`.")
-        values = jnp.arange(total_count + 1).reshape((-1,) + (1,) * len(self.batch_shape))
-        if expand:
-            values = jnp.broadcast_to(values, values.shape[:1] + self.batch_shape)
-        return values
-
 
 def Binomial(total_count=1, probs=None, logits=None, validate_args=None):
     if probs is not None:
@@ -268,6 +274,7 @@ class CategoricalProbs(Distribution):
                                                validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return categorical(key, self.probs, shape=sample_shape + self.batch_shape)
 
     @validate_sample
@@ -275,10 +282,14 @@ class CategoricalProbs(Distribution):
         batch_shape = lax.broadcast_shapes(jnp.shape(value), self.batch_shape)
         value = jnp.expand_dims(value, axis=-1)
         value = jnp.broadcast_to(value, batch_shape + (1,))
-        logits = _to_logits_multinom(self.probs)
+        logits = self.logits
         log_pmf = jnp.broadcast_to(logits, batch_shape + jnp.shape(logits)[-1:])
         return jnp.take_along_axis(log_pmf, value, axis=-1)[..., 0]
 
+    @lazy_property
+    def logits(self):
+        return _to_logits_multinom(self.probs)
+
     @property
     def mean(self):
         return jnp.full(self.batch_shape, jnp.nan, dtype=get_dtype(self.probs))
@@ -311,6 +322,7 @@ class CategoricalLogits(Distribution):
                                                 validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return random.categorical(key, self.logits, shape=sample_shape + self.batch_shape)
 
     @validate_sample
@@ -354,49 +366,6 @@ def Categorical(probs=None, logits=None, validate_args=None):
         raise ValueError('One of `probs` or `logits` must be specified.')
 
 
-class Delta(Distribution):
-    arg_constraints = {'value': constraints.real, 'log_density': constraints.real}
-    support = constraints.real
-    is_discrete = True
-
-    def __init__(self, value=0., log_density=0., event_dim=0, validate_args=None):
-        if event_dim > jnp.ndim(value):
-            raise ValueError('Expected event_dim <= v.dim(), actual {} vs {}'
-                             .format(event_dim, jnp.ndim(value)))
-        batch_dim = jnp.ndim(value) - event_dim
-        batch_shape = jnp.shape(value)[:batch_dim]
-        event_shape = jnp.shape(value)[batch_dim:]
-        self.value = lax.convert_element_type(value, canonicalize_dtype(jnp.float64))
-        # NB: following Pyro implementation, log_density should be broadcasted to batch_shape
-        self.log_density = promote_shapes(log_density, shape=batch_shape)[0]
-        super(Delta, self).__init__(batch_shape, event_shape, validate_args=validate_args)
-
-    def sample(self, key, sample_shape=()):
-        shape = sample_shape + self.batch_shape + self.event_shape
-        return jnp.broadcast_to(device_put(self.value), shape)
-
-    @validate_sample
-    def log_prob(self, value):
-        log_prob = jnp.log(value == self.value)
-        log_prob = sum_rightmost(log_prob, len(self.event_shape))
-        return log_prob + self.log_density
-
-    @property
-    def mean(self):
-        return self.value
-
-    @property
-    def variance(self):
-        return jnp.zeros(self.batch_shape + self.event_shape)
-
-    def tree_flatten(self):
-        return (self.value, self.log_density), self.event_dim
-
-    @classmethod
-    def tree_unflatten(cls, aux_data, params):
-        return cls(*params, event_dim=aux_data)
-
-
 class OrderedLogistic(CategoricalProbs):
     """
     A categorical distribution with ordered outcomes.
@@ -414,7 +383,11 @@ class OrderedLogistic(CategoricalProbs):
                        'cutpoints': constraints.ordered_vector}
 
     def __init__(self, predictor, cutpoints, validate_args=None):
-        predictor, self.cutpoints = promote_shapes(jnp.expand_dims(predictor, -1), cutpoints)
+        if jnp.ndim(predictor) == 0:
+            predictor, = promote_shapes(predictor, shape=(1,))
+        else:
+            predictor = predictor[..., None]
+        predictor, self.cutpoints = promote_shapes(predictor, cutpoints)
         self.predictor = predictor[..., 0]
         cumulative_probs = expit(cutpoints - predictor)
         # add two boundary points 0 and 1
@@ -433,6 +406,9 @@ class PRNGIdentity(Distribution):
     is_discrete = True
 
     def __init__(self):
+        warnings.warn("PRNGIdentity distribution is deprecated. To get a random "
+                      "PRNG key, you can use `numpyro.prng_key()` instead.",
+                      FutureWarning)
         super(PRNGIdentity, self).__init__(event_shape=(2,))
 
     def sample(self, key, sample_shape=()):
@@ -456,6 +432,7 @@ class MultinomialProbs(Distribution):
                                                validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return multinomial(key, self.probs, self.total_count, shape=sample_shape + self.batch_shape)
 
     @validate_sample
@@ -465,6 +442,10 @@ class MultinomialProbs(Distribution):
         return gammaln(self.total_count + 1) \
             + jnp.sum(xlogy(value, self.probs) - gammaln(value + 1), axis=-1)
 
+    @lazy_property
+    def logits(self):
+        return _to_logits_multinom(self.probs)
+
     @property
     def mean(self):
         return self.probs * jnp.expand_dims(self.total_count, -1)
@@ -494,6 +475,7 @@ class MultinomialLogits(Distribution):
                                                 validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return multinomial(key, self.probs, self.total_count, shape=sample_shape + self.batch_shape)
 
     @validate_sample
@@ -540,6 +522,7 @@ class Poisson(Distribution):
         super(Poisson, self).__init__(jnp.shape(rate), validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         return random.poisson(key, self.rate, shape=sample_shape + self.batch_shape)
 
     @validate_sample
@@ -574,6 +557,7 @@ class ZeroInflatedPoisson(Distribution):
         super(ZeroInflatedPoisson, self).__init__(batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
         key_bern, key_poisson = random.split(key)
         shape = sample_shape + self.batch_shape
         mask = random.bernoulli(key_bern, self.gate, shape)
@@ -592,3 +576,83 @@ class ZeroInflatedPoisson(Distribution):
     @lazy_property
     def variance(self):
         return (1 - self.gate) * self.rate * (1 + self.rate * self.gate)
+
+
+class GeometricProbs(Distribution):
+    arg_constraints = {'probs': constraints.unit_interval}
+    support = constraints.nonnegative_integer
+    is_discrete = True
+
+    def __init__(self, probs, validate_args=None):
+        self.probs = probs
+        super(GeometricProbs, self).__init__(batch_shape=jnp.shape(self.probs),
+                                             validate_args=validate_args)
+
+    def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
+        probs = self.probs
+        dtype = get_dtype(probs)
+        shape = sample_shape + self.batch_shape
+        u = random.uniform(key, shape, dtype)
+        return jnp.floor(jnp.log1p(-u) / jnp.log1p(-probs))
+
+    @validate_sample
+    def log_prob(self, value):
+        probs = jnp.where((self.probs == 1) & (value == 0), 0, self.probs)
+        return value * jnp.log1p(-probs) + jnp.log(probs)
+
+    @lazy_property
+    def logits(self):
+        return _to_logits_bernoulli(self.probs)
+
+    @property
+    def mean(self):
+        return 1. / self.probs - 1.
+
+    @property
+    def variance(self):
+        return (1. / self.probs - 1.) / self.probs
+
+
+class GeometricLogits(Distribution):
+    arg_constraints = {'logits': constraints.real}
+    support = constraints.nonnegative_integer
+    is_discrete = True
+
+    def __init__(self, logits, validate_args=None):
+        self.logits = logits
+        super(GeometricLogits, self).__init__(batch_shape=jnp.shape(self.logits),
+                                              validate_args=validate_args)
+
+    @lazy_property
+    def probs(self):
+        return _to_probs_bernoulli(self.logits)
+
+    def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
+        logits = self.logits
+        dtype = get_dtype(logits)
+        shape = sample_shape + self.batch_shape
+        u = random.uniform(key, shape, dtype)
+        return jnp.floor(jnp.log1p(-u) / -softplus(logits))
+
+    @validate_sample
+    def log_prob(self, value):
+        return (-value - 1) * softplus(self.logits) + self.logits
+
+    @property
+    def mean(self):
+        return 1. / self.probs - 1.
+
+    @property
+    def variance(self):
+        return (1. / self.probs - 1.) / self.probs
+
+
+def Geometric(probs=None, logits=None, validate_args=None):
+    if probs is not None:
+        return GeometricProbs(probs, validate_args=validate_args)
+    elif logits is not None:
+        return GeometricLogits(logits, validate_args=validate_args)
+    else:
+        raise ValueError('One of `probs` or `logits` must be specified.')
