@@ -27,13 +27,16 @@
 
 from collections import OrderedDict
 from contextlib import contextmanager
+import functools
 import warnings
 
+import numpy as np
+
 from jax import lax, tree_util
 import jax.numpy as jnp
 
-from numpyro.distributions.constraints import is_dependent, real
-from numpyro.distributions.transforms import Transform
+from numpyro.distributions.constraints import independent, is_dependent, real
+from numpyro.distributions.transforms import ComposeTransform, Transform
 from numpyro.distributions.util import lazy_property, promote_shapes, sum_rightmost, validate_sample
 from numpyro.util import not_jax_tracer
 
@@ -71,7 +74,23 @@ def validation_enabled(is_validate=True):
         enable_validation(distribution_validation_status)
 
 
-class Distribution(object):
+COERCIONS = []
+
+
+class DistributionMeta(type):
+    def __call__(cls, *args, **kwargs):
+        for coerce_ in COERCIONS:
+            result = coerce_(cls, args, kwargs)
+            if result is not None:
+                return result
+        return super().__call__(*args, **kwargs)
+
+    @property
+    def __wrapped__(cls):
+        return functools.partial(cls.__init__, None)
+
+
+class Distribution(metaclass=DistributionMeta):
     """
     Base class for probability distributions in NumPyro. The design largely
     follows from :mod:`torch.distributions`.
@@ -138,10 +157,11 @@ class Distribution(object):
                     continue
                 if is_dependent(constraint):
                     continue  # skip constraints that cannot be checked
-                is_valid = jnp.all(constraint(getattr(self, param)))
+                is_valid = constraint(getattr(self, param))
                 if not_jax_tracer(is_valid):
-                    if not is_valid:
-                        raise ValueError("The parameter {} has invalid values".format(param))
+                    if not np.all(is_valid):
+                        raise ValueError("{} distribution got invalid {} parameter.".format(
+                            self.__class__.__name__, param))
         super(Distribution, self).__init__()
 
     @property
@@ -173,6 +193,16 @@ class Distribution(object):
         """
         return len(self.event_shape)
 
+    @property
+    def has_rsample(self):
+        return set(self.reparametrized_params) == set(self.arg_constraints)
+
+    def rsample(self, key, sample_shape=()):
+        if self.has_rsample:
+            return self.sample(key, sample_shape=sample_shape)
+
+        raise NotImplementedError
+
     def shape(self, sample_shape=()):
         """
         The tensor shape of samples from this distribution.
@@ -242,7 +272,7 @@ class Distribution(object):
     def _validate_sample(self, value):
         mask = self.support(value)
         if not_jax_tracer(mask):
-            if not jnp.all(mask):
+            if not np.all(mask):
                 warnings.warn('Out-of-support values provided to log prob method. '
                               'The value argument should be within the support.')
         return mask
@@ -262,11 +292,11 @@ class Distribution(object):
         :param reinterpreted_batch_ndims: Number of rightmost batch dims to
             interpret as event dims.
         :return: An instance of `Independent` distribution.
-        :rtype: Independent
+        :rtype: numpyro.distributions.distribution.Independent
         """
         if reinterpreted_batch_ndims is None:
             reinterpreted_batch_ndims = len(self.batch_shape)
-        elif reinterpreted_batch_ndims == 0:
+        if reinterpreted_batch_ndims == 0:
             return self
         return Independent(self, reinterpreted_batch_ndims)
 
@@ -303,7 +333,7 @@ class Distribution(object):
         :return: An expanded version of this distribution.
         :rtype: :class:`ExpandedDistribution`
         """
-        return self.expand(tuple(sample_shape) + self._batch_shape)
+        return self.expand(tuple(sample_shape) + self.batch_shape)
 
     def mask(self, mask):
         """
@@ -327,23 +357,19 @@ class ExpandedDistribution(Distribution):
 
     def __init__(self, base_dist, batch_shape=()):
         if isinstance(base_dist, ExpandedDistribution):
-            batch_shape = self._broadcast_shape(base_dist.batch_shape, batch_shape)
+            batch_shape, _, _ = self._broadcast_shape(base_dist.batch_shape, batch_shape)
             base_dist = base_dist.base_dist
         self.base_dist = base_dist
-        super().__init__(base_dist.batch_shape, base_dist.event_shape)
-        # adjust batch shape
-        self.expand(batch_shape)
 
-    def expand(self, batch_shape):
+        # adjust batch shape
         # Do basic validation. e.g. we should not "unexpand" distributions even if that is possible.
-        new_shape, _, _ = self._broadcast_shape(self.batch_shape, batch_shape)
+        new_shape, _, _ = self._broadcast_shape(base_dist.batch_shape, batch_shape)
         # Record interstitial and expanded dims/sizes w.r.t. the base distribution
-        new_shape, expanded_sizes, interstitial_sizes = self._broadcast_shape(self.base_dist.batch_shape,
+        new_shape, expanded_sizes, interstitial_sizes = self._broadcast_shape(base_dist.batch_shape,
                                                                               new_shape)
-        self._batch_shape = new_shape
         self._expanded_sizes = expanded_sizes
         self._interstitial_sizes = interstitial_sizes
-        return self
+        super().__init__(new_shape, base_dist.event_shape)
 
     @staticmethod
     def _broadcast_shape(existing_shape, new_shape):
@@ -374,23 +400,33 @@ class ExpandedDistribution(Distribution):
         return self.base_dist.is_discrete
 
     @property
-    def support(self):
-        return self.base_dist.support
+    def has_rsample(self):
+        return self.base_dist.has_rsample
 
-    def sample(self, key, sample_shape=()):
+    def _sample(self, sample_fn, key, sample_shape=()):
         interstitial_dims = tuple(self._interstitial_sizes.keys())
         event_dim = len(self.event_shape)
         interstitial_dims = tuple(i - event_dim for i in interstitial_dims)
         interstitial_sizes = tuple(self._interstitial_sizes.values())
         expanded_sizes = tuple(self._expanded_sizes.values())
         batch_shape = expanded_sizes + interstitial_sizes
-        samples = self.base_dist.sample(key, sample_shape + batch_shape)
+        samples = sample_fn(key, sample_shape=sample_shape + batch_shape)
         interstitial_idx = len(sample_shape) + len(expanded_sizes)
         interstitial_sample_dims = tuple(range(interstitial_idx, interstitial_idx + len(interstitial_sizes)))
         for dim1, dim2 in zip(interstitial_dims, interstitial_sample_dims):
             samples = jnp.swapaxes(samples, dim1, dim2)
         return samples.reshape(sample_shape + self.batch_shape + self.event_shape)
 
+    def rsample(self, key, sample_shape=()):
+        return self._sample(self.base_dist.rsample, key, sample_shape)
+
+    @property
+    def support(self):
+        return self.base_dist.support
+
+    def sample(self, key, sample_shape=()):
+        return self._sample(self.base_dist.sample, key, sample_shape)
+
     def log_prob(self, value):
         shape = lax.broadcast_shapes(self.batch_shape,
                                      jnp.shape(value)[:max(jnp.ndim(value) - self.event_dim, 0)])
@@ -482,7 +518,7 @@ class ImproperUniform(Distribution):
     arg_constraints = {}
 
     def __init__(self, support, batch_shape, event_shape, validate_args=None):
-        self.support = support
+        self.support = independent(support, len(event_shape) - support.event_dim)
         super().__init__(batch_shape, event_shape, validate_args=validate_args)
 
     @validate_sample
@@ -546,7 +582,7 @@ class Independent(Distribution):
 
     @property
     def support(self):
-        return self.base_dist.support
+        return independent(self.base_dist.support, self.reinterpreted_batch_ndims)
 
     @property
     def has_enumerate_support(self):
@@ -568,8 +604,15 @@ class Independent(Distribution):
     def variance(self):
         return self.base_dist.variance
 
+    @property
+    def has_rsample(self):
+        return self.base_dist.has_rsample
+
+    def rsample(self, key, sample_shape=()):
+        return self.base_dist.rsample(key, sample_shape=sample_shape)
+
     def sample(self, key, sample_shape=()):
-        return self.base_dist.sample(key, sample_shape=sample_shape)
+        return self.base_dist(rng_key=key, sample_shape=sample_shape)
 
     def log_prob(self, value):
         log_prob = self.base_dist.log_prob(value)
@@ -606,7 +649,7 @@ class MaskedDistribution(Distribution):
         if isinstance(mask, bool):
             self._mask = mask
         else:
-            batch_shape = lax.broadcast_shapes(jnp.shape(mask), base_dist.batch_shape)
+            batch_shape = lax.broadcast_shapes(jnp.shape(mask), tuple(base_dist.batch_shape))
             if mask.shape != batch_shape:
                 mask = jnp.broadcast_to(mask, batch_shape)
             if base_dist.batch_shape != batch_shape:
@@ -623,20 +666,34 @@ class MaskedDistribution(Distribution):
     def is_discrete(self):
         return self.base_dist.is_discrete
 
+    @property
+    def has_rsample(self):
+        return self.base_dist.has_rsample
+
+    def rsample(self, key, sample_shape=()):
+        return self.base_dist.rsample(key, sample_shape=sample_shape)
+
     @property
     def support(self):
         return self.base_dist.support
 
     def sample(self, key, sample_shape=()):
-        return self.base_dist.sample(key, sample_shape)
+        return self.base_dist(rng_key=key, sample_shape=sample_shape)
 
     def log_prob(self, value):
         if self._mask is False:
-            shape = lax.broadcast_shapes(self.base_dist.batch_shape,
+            shape = lax.broadcast_shapes(tuple(self.base_dist.batch_shape),
                                          jnp.shape(value)[:max(jnp.ndim(value) - len(self.event_shape), 0)])
             return jnp.zeros(shape)
         if self._mask is True:
             return self.base_dist.log_prob(value)
+        try:
+            default_value = self.base_dist.support.feasible_like(value)
+        except NotImplementedError:
+            pass
+        else:
+            mask = jnp.reshape(self._mask, jnp.shape(self._mask) + (1,) * self.event_dim)
+            value = jnp.where(mask, value, default_value)
         return jnp.where(self._mask, self.base_dist.log_prob(value), 0.)
 
     def enumerate_support(self, expand=True):
@@ -691,45 +748,65 @@ class TransformedDistribution(Distribution):
                 raise ValueError("transforms must be a Transform or a list of Transforms")
         else:
             raise ValueError("transforms must be a Transform or list, but was {}".format(transforms))
-        # XXX: this logic will not be valid when IndependentDistribution is support;
-        # in that case, it is more involved to support Transform(Indep(Transform));
-        # however, we might not need to support such kind of distribution
-        # and should raise an error if base_distribution is an Indep one
         if isinstance(base_distribution, TransformedDistribution):
-            self.base_dist = base_distribution.base_dist
+            base_dist = base_distribution.base_dist
             self.transforms = base_distribution.transforms + transforms
         else:
-            self.base_dist = base_distribution
+            base_dist = base_distribution
             self.transforms = transforms
-        # NB: here we assume that base_dist.shape == transformed_dist.shape
-        # but that might not be True for some transforms such as StickBreakingTransform
-        # because the event dimension is transformed from (n - 1,) to (n,).
-        # Currently, we have no mechanism to fix this issue. Given that
-        # this is just an edge case, we might skip this issue but need
-        # to pay attention to any inference function that inspects
-        # transformed distribution's shape.
-        shape = base_distribution.batch_shape + base_distribution.event_shape
-        event_dim = max([len(base_distribution.event_shape)] + [t.event_dim for t in transforms])
-        batch_shape = shape[:len(shape) - event_dim]
-        event_shape = shape[len(shape) - event_dim:]
+        base_shape = base_dist.shape()
+        base_event_dim = base_dist.event_dim
+        transform = ComposeTransform(self.transforms)
+        domain_event_dim = transform.domain.event_dim
+        if len(base_shape) < domain_event_dim:
+            raise ValueError("Base distribution needs to have shape with size at least {}, but got {}."
+                             .format(domain_event_dim, base_shape))
+        shape = transform.forward_shape(base_shape)
+        expanded_base_shape = transform.inverse_shape(shape)
+        if base_shape != expanded_base_shape:
+            base_batch_shape = expanded_base_shape[:len(expanded_base_shape) - base_event_dim]
+            base_dist = base_dist.expand(base_batch_shape)
+        reinterpreted_batch_ndims = domain_event_dim - base_event_dim
+        if reinterpreted_batch_ndims > 0:
+            base_dist = base_dist.to_event(reinterpreted_batch_ndims)
+        self.base_dist = base_dist
+
+        # Compute shapes.
+        event_dim = transform.codomain.event_dim + max(base_event_dim - domain_event_dim, 0)
+        assert len(shape) >= event_dim
+        cut = len(shape) - event_dim
+        batch_shape = shape[:cut]
+        event_shape = shape[cut:]
         super(TransformedDistribution, self).__init__(batch_shape, event_shape, validate_args=validate_args)
 
+    @property
+    def has_rsample(self):
+        return self.base_dist.has_rsample
+
+    def rsample(self, key, sample_shape=()):
+        x = self.base_dist.rsample(key, sample_shape=sample_shape)
+        for transform in self.transforms:
+            x = transform(x)
+        return x
+
     @property
     def support(self):
-        domain = self.base_dist.support
-        for t in self.transforms:
-            t.domain = domain
-            domain = t.codomain
-        return domain
+        codomain = self.transforms[-1].codomain
+        codomain_event_dim = codomain.event_dim
+        assert self.event_dim >= codomain_event_dim
+        if self.event_dim == codomain_event_dim:
+            return codomain
+        else:
+            return independent(codomain, self.event_dim - codomain_event_dim)
 
     def sample(self, key, sample_shape=()):
-        x = self.base_dist.sample(key, sample_shape)
+        x = self.base_dist(rng_key=key, sample_shape=sample_shape)
         for transform in self.transforms:
             x = transform(x)
         return x
 
     def sample_with_intermediates(self, key, sample_shape=()):
-        x = self.base_dist.sample(key, sample_shape)
+        x = self.base_dist(rng_key=key, sample_shape=sample_shape)
         intermediates = []
         for transform in self.transforms:
             x_tmp = x
@@ -750,7 +827,9 @@ class TransformedDistribution(Distribution):
             x = transform.inv(y) if intermediates is None else intermediates[-i - 1][0]
             t_inter = None if intermediates is None else intermediates[-i - 1][1]
             t_log_det = transform.log_abs_det_jacobian(x, y, t_inter)
-            log_prob = log_prob - sum_rightmost(t_log_det, event_dim - transform.event_dim)
+            batch_ndim = event_dim - transform.codomain.event_dim
+            log_prob = log_prob - sum_rightmost(t_log_det, batch_ndim)
+            event_dim = transform.domain.event_dim + batch_ndim
             y = x
 
         log_prob = log_prob + sum_rightmost(self.base_dist.log_prob(y),
@@ -773,6 +852,53 @@ class TransformedDistribution(Distribution):
             " your usage cases.")
 
 
+class Delta(Distribution):
+    arg_constraints = {'v': real, 'log_density': real}
+    reparameterized_params = ['v', 'log_density']
+    is_discrete = True
+
+    def __init__(self, v=0., log_density=0., event_dim=0, validate_args=None):
+        if event_dim > jnp.ndim(v):
+            raise ValueError('Expected event_dim <= v.dim(), actual {} vs {}'
+                             .format(event_dim, jnp.ndim(v)))
+        batch_dim = jnp.ndim(v) - event_dim
+        batch_shape = jnp.shape(v)[:batch_dim]
+        event_shape = jnp.shape(v)[batch_dim:]
+        self.v = v
+        # NB: following Pyro implementation, log_density should be broadcasted to batch_shape
+        self.log_density = promote_shapes(log_density, shape=batch_shape)[0]
+        super(Delta, self).__init__(batch_shape, event_shape, validate_args=validate_args)
+
+    @property
+    def support(self):
+        return independent(real, self.event_dim)
+
+    def sample(self, key, sample_shape=()):
+        shape = sample_shape + self.batch_shape + self.event_shape
+        return jnp.broadcast_to(self.v, shape)
+
+    @validate_sample
+    def log_prob(self, value):
+        log_prob = jnp.log(value == self.v)
+        log_prob = sum_rightmost(log_prob, len(self.event_shape))
+        return log_prob + self.log_density
+
+    @property
+    def mean(self):
+        return self.v
+
+    @property
+    def variance(self):
+        return jnp.zeros(self.batch_shape + self.event_shape)
+
+    def tree_flatten(self):
+        return (self.v, self.log_density), self.event_dim
+
+    @classmethod
+    def tree_unflatten(cls, aux_data, params):
+        return cls(*params, event_dim=aux_data)
+
+
 class Unit(Distribution):
     """
     Trivial nonnormalized distribution representing the unit type.
