@@ -3,8 +3,11 @@
 
 import math
 import warnings
+import weakref
 
-from jax import ops, tree_flatten, tree_map, vmap
+import numpy as np
+
+from jax import lax, ops, tree_flatten, tree_map, vmap
 from jax.dtypes import canonicalize_dtype
 from jax.flatten_util import ravel_pytree
 from jax.nn import softplus
@@ -50,12 +53,29 @@ def _clipped_expit(x):
 class Transform(object):
     domain = constraints.real
     codomain = constraints.real
-    event_dim = 0
+    _inv = None
+
+    @property
+    def event_dim(self):
+        warnings.warn("transform.event_dim is deprecated. Please use Transform.domain.event_dim to "
+                      "get input event dim or Transform.codomain.event_dim to get output event dim.",
+                      FutureWarning)
+        return self.domain.event_dim
+
+    @property
+    def inv(self):
+        inv = None
+        if self._inv is not None:
+            inv = self._inv()
+        if inv is None:
+            inv = _InverseTransform(self)
+            self._inv = weakref.ref(inv)
+        return inv
 
     def __call__(self, x):
         return NotImplementedError
 
-    def inv(self, y):
+    def _inverse(self, y):
         raise NotImplementedError
 
     def log_abs_det_jacobian(self, x, y, intermediates=None):
@@ -64,6 +84,51 @@ class Transform(object):
     def call_with_intermediates(self, x):
         return self(x), None
 
+    def forward_shape(self, shape):
+        """
+        Infers the shape of the forward computation, given the input shape.
+        Defaults to preserving shape.
+        """
+        return shape
+
+    def inverse_shape(self, shape):
+        """
+        Infers the shapes of the inverse computation, given the output shape.
+        Defaults to preserving shape.
+        """
+        return shape
+
+
+class _InverseTransform(Transform):
+    def __init__(self, transform):
+        super().__init__()
+        self._inv = transform
+
+    @property
+    def domain(self):
+        return self._inv.codomain
+
+    @property
+    def codomain(self):
+        return self._inv.domain
+
+    @property
+    def inv(self):
+        return self._inv
+
+    def __call__(self, x):
+        return self._inv._inverse(x)
+
+    def log_abs_det_jacobian(self, x, y, intermediates=None):
+        # NB: we don't use intermediates for inverse transform
+        return -self._inv.log_abs_det_jacobian(y, x, None)
+
+    def forward_shape(self, shape):
+        return self._inv.inverse_shape(shape)
+
+    def inverse_shape(self, shape):
+        return self._inv.forward_shape(shape)
+
 
 class AbsTransform(Transform):
     domain = constraints.real
@@ -75,7 +140,7 @@ class AbsTransform(Transform):
     def __call__(self, x):
         return jnp.abs(x)
 
-    def inv(self, y):
+    def _inverse(self, y):
         return y
 
 
@@ -93,22 +158,20 @@ class AffineTransform(Transform):
     def codomain(self):
         if self.domain is constraints.real:
             return constraints.real
-        elif self.domain is constraints.real_vector:
-            return constraints.real_vector
         elif isinstance(self.domain, constraints.greater_than):
-            if not_jax_tracer(self.scale) and jnp.all(self.scale < 0):
+            if not_jax_tracer(self.scale) and np.all(np.less(self.scale, 0)):
                 return constraints.less_than(self(self.domain.lower_bound))
             # we suppose scale > 0 for any tracer
             else:
                 return constraints.greater_than(self(self.domain.lower_bound))
         elif isinstance(self.domain, constraints.less_than):
-            if not_jax_tracer(self.scale) and jnp.all(self.scale < 0):
+            if not_jax_tracer(self.scale) and np.all(np.less(self.scale, 0)):
                 return constraints.greater_than(self(self.domain.upper_bound))
             # we suppose scale > 0 for any tracer
             else:
                 return constraints.less_than(self(self.domain.upper_bound))
         elif isinstance(self.domain, constraints.interval):
-            if not_jax_tracer(self.scale) and jnp.all(self.scale < 0):
+            if not_jax_tracer(self.scale) and np.all(np.less(self.scale, 0)):
                 return constraints.interval(self(self.domain.upper_bound),
                                             self(self.domain.lower_bound))
             else:
@@ -117,18 +180,38 @@ class AffineTransform(Transform):
         else:
             raise NotImplementedError
 
-    @property
-    def event_dim(self):
-        return 1 if self.domain is constraints.real_vector else 0
-
     def __call__(self, x):
         return self.loc + self.scale * x
 
-    def inv(self, y):
+    def _inverse(self, y):
         return (y - self.loc) / self.scale
 
     def log_abs_det_jacobian(self, x, y, intermediates=None):
-        return sum_rightmost(jnp.broadcast_to(jnp.log(jnp.abs(self.scale)), jnp.shape(x)), self.event_dim)
+        return jnp.broadcast_to(jnp.log(jnp.abs(self.scale)), jnp.shape(x))
+
+    def forward_shape(self, shape):
+        return lax.broadcast_shapes(shape,
+                                    getattr(self.loc, "shape", ()),
+                                    getattr(self.scale, "shape", ()))
+
+    def inverse_shape(self, shape):
+        return lax.broadcast_shapes(shape,
+                                    getattr(self.loc, "shape", ()),
+                                    getattr(self.scale, "shape", ()))
+
+
+def _get_compose_transform_input_event_dim(parts):
+    input_event_dim = parts[-1].domain.event_dim
+    for part in parts[len(parts) - 1::-1]:
+        input_event_dim = part.domain.event_dim + max(input_event_dim - part.codomain.event_dim, 0)
+    return input_event_dim
+
+
+def _get_compose_transform_output_event_dim(parts):
+    output_event_dim = parts[0].codomain.event_dim
+    for part in parts[1:]:
+        output_event_dim = part.codomain.event_dim + max(output_event_dim - part.domain.event_dim, 0)
+    return output_event_dim
 
 
 class ComposeTransform(Transform):
@@ -137,22 +220,30 @@ class ComposeTransform(Transform):
 
     @property
     def domain(self):
-        return self.parts[0].domain
+        input_event_dim = _get_compose_transform_input_event_dim(self.parts)
+        first_input_event_dim = self.parts[0].domain.event_dim
+        assert input_event_dim >= first_input_event_dim
+        if input_event_dim == first_input_event_dim:
+            return self.parts[0].domain
+        else:
+            return constraints.independent(self.parts[0].domain, input_event_dim - first_input_event_dim)
 
     @property
     def codomain(self):
-        return self.parts[-1].codomain
-
-    @property
-    def event_dim(self):
-        return max(p.event_dim for p in self.parts)
+        output_event_dim = _get_compose_transform_output_event_dim(self.parts)
+        last_output_event_dim = self.parts[-1].codomain.event_dim
+        assert output_event_dim >= last_output_event_dim
+        if output_event_dim == last_output_event_dim:
+            return self.parts[-1].codomain
+        else:
+            return constraints.independent(self.parts[-1].codomain, output_event_dim - last_output_event_dim)
 
     def __call__(self, x):
         for part in self.parts:
             x = part(x)
         return x
 
-    def inv(self, y):
+    def _inverse(self, y):
         for part in self.parts[::-1]:
             y = part.inv(y)
         return y
@@ -164,16 +255,20 @@ class ComposeTransform(Transform):
                                  .format(len(intermediates), len(self.parts)))
 
         result = 0.
+        input_event_dim = self.domain.event_dim
         for i, part in enumerate(self.parts[:-1]):
             y_tmp = part(x) if intermediates is None else intermediates[i][0]
             inter = None if intermediates is None else intermediates[i][1]
             logdet = part.log_abs_det_jacobian(x, y_tmp, intermediates=inter)
-            result = result + sum_rightmost(logdet, self.event_dim - part.event_dim)
+            batch_ndim = input_event_dim - part.domain.event_dim
+            result = result + sum_rightmost(logdet, batch_ndim)
+            input_event_dim = part.codomain.event_dim + batch_ndim
             x = y_tmp
         # account the the last transform, where y is available
         inter = None if intermediates is None else intermediates[-1]
-        logdet = self.parts[-1].log_abs_det_jacobian(x, y, intermediates=inter)
-        result = result + sum_rightmost(logdet, self.event_dim - self.parts[-1].event_dim)
+        part = self.parts[-1]
+        logdet = part.log_abs_det_jacobian(x, y, intermediates=inter)
+        result = result + sum_rightmost(logdet, input_event_dim - part.domain.event_dim)
         return result
 
     def call_with_intermediates(self, x):
@@ -186,6 +281,39 @@ class ComposeTransform(Transform):
         intermediates.append(inter)
         return x, intermediates
 
+    def forward_shape(self, shape):
+        for part in self.parts:
+            shape = part.forward_shape(shape)
+        return shape
+
+    def inverse_shape(self, shape):
+        for part in reversed(self.parts):
+            shape = part.inverse_shape(shape)
+        return shape
+
+
+def _matrix_forward_shape(shape, offset=0):
+    # Reshape from (..., N) to (..., D, D).
+    if len(shape) < 1:
+        raise ValueError("Too few dimensions in input")
+    N = shape[-1]
+    D = round((0.25 + 2 * N) ** 0.5 - 0.5)
+    if D * (D + 1) // 2 != N:
+        raise ValueError("Input is not a flattend lower-diagonal number")
+    D = D - offset
+    return shape[:-1] + (D, D)
+
+
+def _matrix_inverse_shape(shape, offset=0):
+    # Reshape from (..., D, D) to (..., N).
+    if len(shape) < 2:
+        raise ValueError("Too few dimensions on input")
+    if shape[-2] != shape[-1]:
+        raise ValueError("Input is not square")
+    D = shape[-1] + offset
+    N = D * (D + 1) // 2
+    return shape[:-2] + (N,)
+
 
 class CorrCholeskyTransform(Transform):
     r"""
@@ -215,14 +343,13 @@ class CorrCholeskyTransform(Transform):
     """
     domain = constraints.real_vector
     codomain = constraints.corr_cholesky
-    event_dim = 2
 
     def __call__(self, x):
         # we interchange step 1 and step 2.a for a better performance
         t = jnp.tanh(x)
         return signed_stick_breaking_tril(t)
 
-    def inv(self, y):
+    def _inverse(self, y):
         # inverse stick-breaking
         z1m_cumprod = 1 - jnp.cumsum(y * y, axis=-1)
         pad_width = [(0, 0)] * y.ndim
@@ -250,6 +377,12 @@ class CorrCholeskyTransform(Transform):
         tanh_logdet = -2 * jnp.sum(x + softplus(-2 * x) - jnp.log(2.), axis=-1)
         return stick_breaking_logdet + tanh_logdet
 
+    def forward_shape(self, shape):
+        return _matrix_forward_shape(shape, offset=-1)
+
+    def inverse_shape(self, shape):
+        return _matrix_inverse_shape(shape, offset=-1)
+
 
 class ExpTransform(Transform):
     # TODO: refine domain/codomain logic through setters, especially when
@@ -273,7 +406,7 @@ class ExpTransform(Transform):
         # XXX consider to clamp from below for stability if necessary
         return jnp.exp(x)
 
-    def inv(self, y):
+    def _inverse(self, y):
         return jnp.log(y)
 
     def log_abs_det_jacobian(self, x, y, intermediates=None):
@@ -282,17 +415,59 @@ class ExpTransform(Transform):
 
 class IdentityTransform(Transform):
 
-    def __init__(self, event_dim=0):
-        self.event_dim = event_dim
-
     def __call__(self, x):
         return x
 
-    def inv(self, y):
+    def _inverse(self, y):
         return y
 
     def log_abs_det_jacobian(self, x, y, intermediates=None):
-        return jnp.full(jnp.shape(x) if self.event_dim == 0 else jnp.shape(x)[:-1], 0.)
+        return jnp.zeros_like(x)
+
+
+class IndependentTransform(Transform):
+    """
+    Wraps a transform by aggregating over ``reinterpreted_batch_ndims``-many
+    dims in :meth:`check`, so that an event is valid only if all its
+    independent entries are valid.
+    """
+    def __init__(self, base_transform, reinterpreted_batch_ndims):
+        assert isinstance(base_transform, Transform)
+        assert isinstance(reinterpreted_batch_ndims, int)
+        assert reinterpreted_batch_ndims >= 0
+        self.base_transform = base_transform
+        self.reinterpreted_batch_ndims = reinterpreted_batch_ndims
+        super().__init__()
+
+    @property
+    def domain(self):
+        return constraints.independent(self.base_transform.domain, self.reinterpreted_batch_ndims)
+
+    @property
+    def codomain(self):
+        return constraints.independent(self.base_transform.codomain, self.reinterpreted_batch_ndims)
+
+    def __call__(self, x):
+        return self.base_transform(x)
+
+    def _inverse(self, y):
+        return self.base_transform._inverse(y)
+
+    def log_abs_det_jacobian(self, x, y, intermediates=None):
+        result = self.base_transform.log_abs_det_jacobian(x, y, intermediates=intermediates)
+        if jnp.ndim(result) < self.reinterpreted_batch_ndims:
+            expected = self.domain.event_dim
+            raise ValueError(f"Expected x.dim() >= {expected} but got {jnp.ndim(x)}")
+        return sum_rightmost(result, self.reinterpreted_batch_ndims)
+
+    def call_with_intermediates(self, x):
+        return self.base_transform.call_with_intermediates(x)
+
+    def forward_shape(self, shape):
+        return self.base_transform.forward_shape(shape)
+
+    def inverse_shape(self, shape):
+        return self.base_transform.inverse_shape(shape)
 
 
 class InvCholeskyTransform(Transform):
@@ -300,7 +475,6 @@ class InvCholeskyTransform(Transform):
     Transform via the mapping :math:`y = x @ x.T`, where `x` is a lower
     triangular matrix with positive diagonal.
     """
-    event_dim = 2
 
     def __init__(self, domain=constraints.lower_cholesky):
         assert domain in [constraints.lower_cholesky, constraints.corr_cholesky]
@@ -310,13 +484,13 @@ class InvCholeskyTransform(Transform):
     def codomain(self):
         if self.domain is constraints.lower_cholesky:
             return constraints.positive_definite
-        elif self.domain:
+        elif self.domain is constraints.corr_cholesky:
             return constraints.corr_matrix
 
     def __call__(self, x):
         return jnp.matmul(x, jnp.swapaxes(x, -2, -1))
 
-    def inv(self, y):
+    def _inverse(self, y):
         return jnp.linalg.cholesky(y)
 
     def log_abs_det_jacobian(self, x, y, intermediates=None):
@@ -341,7 +515,6 @@ class LowerCholeskyAffine(Transform):
     """
     domain = constraints.real_vector
     codomain = constraints.real_vector
-    event_dim = 1
 
     def __init__(self, loc, scale_tril):
         if jnp.ndim(scale_tril) != 2:
@@ -354,7 +527,7 @@ class LowerCholeskyAffine(Transform):
     def __call__(self, x):
         return self.loc + jnp.squeeze(jnp.matmul(self.scale_tril, x[..., jnp.newaxis]), axis=-1)
 
-    def inv(self, y):
+    def _inverse(self, y):
         y = y - self.loc
         original_shape = jnp.shape(y)
         yt = jnp.reshape(y, (-1, original_shape[-1])).T
@@ -365,11 +538,20 @@ class LowerCholeskyAffine(Transform):
         return jnp.broadcast_to(jnp.log(jnp.diagonal(self.scale_tril, axis1=-2, axis2=-1)).sum(-1),
                                 jnp.shape(x)[:-1])
 
+    def forward_shape(self, shape):
+        if len(shape) < 1:
+            raise ValueError("Too few dimensions on input")
+        return lax.broadcast_shapes(shape, self.loc.shape, self.scale_tril.shape[:-1])
+
+    def inverse_shape(self, shape):
+        if len(shape) < 1:
+            raise ValueError("Too few dimensions on input")
+        return lax.broadcast_shapes(shape, self.loc.shape, self.scale_tril.shape[:-1])
+
 
 class LowerCholeskyTransform(Transform):
     domain = constraints.real_vector
     codomain = constraints.lower_cholesky
-    event_dim = 2
 
     def __call__(self, x):
         n = round((math.sqrt(1 + 8 * x.shape[-1]) - 1) / 2)
@@ -377,7 +559,7 @@ class LowerCholeskyTransform(Transform):
         diag = jnp.exp(x[..., -n:])
         return z + jnp.expand_dims(diag, axis=-1) * jnp.identity(n)
 
-    def inv(self, y):
+    def _inverse(self, y):
         z = matrix_to_tril_vec(y, diagonal=-1)
         return jnp.concatenate([z, jnp.log(jnp.diagonal(y, axis1=-2, axis2=-1))], axis=-1)
 
@@ -386,12 +568,11 @@ class LowerCholeskyTransform(Transform):
         n = round((math.sqrt(1 + 8 * x.shape[-1]) - 1) / 2)
         return x[..., -n:].sum(-1)
 
+    def forward_shape(self, shape):
+        return _matrix_forward_shape(shape)
 
-class MultivariateAffineTransform(LowerCholeskyAffine):
-    def __init__(self, loc, scale_tril):
-        warnings.warn("MultivariateAffineTransform is renamed to LowerCholeskyAffine.",
-                      FutureWarning)
-        super().__init__(loc, scale_tril)
+    def inverse_shape(self, shape):
+        return _matrix_inverse_shape(shape)
 
 
 class OrderedTransform(Transform):
@@ -405,13 +586,12 @@ class OrderedTransform(Transform):
     """
     domain = constraints.real_vector
     codomain = constraints.ordered_vector
-    event_dim = 1
 
     def __call__(self, x):
         z = jnp.concatenate([x[..., :1], jnp.exp(x[..., 1:])], axis=-1)
         return jnp.cumsum(z, axis=-1)
 
-    def inv(self, y):
+    def _inverse(self, y):
         x = jnp.log(y[..., 1:] - y[..., :-1])
         return jnp.concatenate([y[..., :1], x], axis=-1)
 
@@ -422,7 +602,6 @@ class OrderedTransform(Transform):
 class PermuteTransform(Transform):
     domain = constraints.real_vector
     codomain = constraints.real_vector
-    event_dim = 1
 
     def __init__(self, permutation):
         self.permutation = permutation
@@ -430,7 +609,7 @@ class PermuteTransform(Transform):
     def __call__(self, x):
         return x[..., self.permutation]
 
-    def inv(self, y):
+    def _inverse(self, y):
         size = self.permutation.size
         permutation_inv = ops.index_update(jnp.zeros(size, dtype=canonicalize_dtype(jnp.int64)),
                                            self.permutation,
@@ -451,12 +630,18 @@ class PowerTransform(Transform):
     def __call__(self, x):
         return jnp.power(x, self.exponent)
 
-    def inv(self, y):
+    def _inverse(self, y):
         return jnp.power(y, 1 / self.exponent)
 
     def log_abs_det_jacobian(self, x, y, intermediates=None):
         return jnp.log(jnp.abs(self.exponent * y / x))
 
+    def forward_shape(self, shape):
+        return lax.broadcast_shapes(shape, getattr(self.exponent, "shape", ()))
+
+    def inverse_shape(self, shape):
+        return lax.broadcast_shapes(shape, getattr(self.exponent, "shape", ()))
+
 
 class SigmoidTransform(Transform):
     codomain = constraints.unit_interval
@@ -464,7 +649,7 @@ class SigmoidTransform(Transform):
     def __call__(self, x):
         return _clipped_expit(x)
 
-    def inv(self, y):
+    def _inverse(self, y):
         return logit(y)
 
     def log_abs_det_jacobian(self, x, y, intermediates=None):
@@ -475,7 +660,6 @@ class SigmoidTransform(Transform):
 class StickBreakingTransform(Transform):
     domain = constraints.real_vector
     codomain = constraints.simplex
-    event_dim = 1
 
     def __call__(self, x):
         # we shift x to obtain a balanced mapping (0, 0, ..., 0) -> (1/K, 1/K, ..., 1/K)
@@ -491,7 +675,7 @@ class StickBreakingTransform(Transform):
         z1m_cumprod_shifted = jnp.pad(z1m_cumprod, pad_width, mode="constant", constant_values=1.)
         return z_padded * z1m_cumprod_shifted
 
-    def inv(self, y):
+    def _inverse(self, y):
         y_crop = y[..., :-1]
         z1m_cumprod = jnp.clip(1 - jnp.cumsum(y_crop, axis=-1), a_min=jnp.finfo(y.dtype).tiny)
         # hence x = logit(z) = log(z / (1 - z)) = y[::-1] / z1m_cumprod
@@ -507,6 +691,16 @@ class StickBreakingTransform(Transform):
         # the case z ~ 1
         return jnp.sum(jnp.log(y[..., :-1] * z) - x, axis=-1)
 
+    def forward_shape(self, shape):
+        if len(shape) < 1:
+            raise ValueError("Too few dimensions on input")
+        return shape[:-1] + (shape[-1] + 1,)
+
+    def inverse_shape(self, shape):
+        if len(shape) < 1:
+            raise ValueError("Too few dimensions on input")
+        return shape[:-1] + (shape[-1] - 1,)
+
 
 class UnpackTransform(Transform):
     """
@@ -515,7 +709,7 @@ class UnpackTransform(Transform):
     :param unpack_fn: callable used to unpack a contiguous array.
     """
     domain = constraints.real_vector
-    event_dim = 1
+    codomain = constraints.dependent
 
     def __init__(self, unpack_fn):
         self.unpack_fn = unpack_fn
@@ -528,7 +722,7 @@ class UnpackTransform(Transform):
         else:
             return self.unpack_fn(x)
 
-    def inv(self, y):
+    def _inverse(self, y):
         leading_dims = [v.shape[0] if jnp.ndim(v) > 0 else 0
                         for v in tree_flatten(y)[0]]
         d0 = leading_dims[0]
@@ -541,6 +735,12 @@ class UnpackTransform(Transform):
     def log_abs_det_jacobian(self, x, y, intermediates=None):
         return jnp.zeros(jnp.shape(x)[:-1])
 
+    def forward_shape(self, shape):
+        raise NotImplementedError
+
+    def inverse_shape(self, shape):
+        raise NotImplementedError
+
 
 ##########################################################
 # CONSTRAINT_REGISTRY
@@ -562,8 +762,8 @@ class ConstraintRegistry(object):
     def __call__(self, constraint):
         try:
             factory = self._registry[type(constraint)]
-        except KeyError:
-            raise NotImplementedError
+        except KeyError as e:
+            raise NotImplementedError from e
 
         return factory(constraint)
 
@@ -578,7 +778,8 @@ def _transform_to_corr_cholesky(constraint):
 
 @biject_to.register(constraints.corr_matrix)
 def _transform_to_corr_matrix(constraint):
-    return ComposeTransform([CorrCholeskyTransform(), InvCholeskyTransform(domain=constraints.corr_cholesky)])
+    return ComposeTransform([CorrCholeskyTransform(),
+                             InvCholeskyTransform(domain=constraints.corr_cholesky)])
 
 
 @biject_to.register(constraints.greater_than)
@@ -597,6 +798,12 @@ def _transform_to_less_than(constraint):
                                              domain=constraints.positive)])
 
 
+@biject_to.register(constraints.independent)
+def _biject_to_independent(constraint):
+    return IndependentTransform(biject_to(constraint.base_constraint),
+                                constraint.reinterpreted_batch_ndims)
+
+
 @biject_to.register(constraints.interval)
 def _transform_to_interval(constraint):
     if constraint is constraints.unit_interval:
@@ -627,11 +834,6 @@ def _transform_to_real(constraint):
     return IdentityTransform()
 
 
-@biject_to.register(constraints.real_vector)
-def _transform_to_real_vector(constraint):
-    return IdentityTransform(event_dim=1)
-
-
 @biject_to.register(constraints.simplex)
 def _transform_to_simplex(constraint):
     return StickBreakingTransform()
