@@ -1,9 +1,12 @@
 # Copyright Contributors to the Pyro project.
 # SPDX-License-Identifier: Apache-2.0
+
 from collections import namedtuple
 from functools import update_wrapper
 import math
 
+import numpy as np
+
 from jax import jit, lax, random, vmap
 from jax.dtypes import canonicalize_dtype
 from jax.lib import xla_bridge
@@ -232,6 +235,13 @@ def binary_cross_entropy_with_logits(x, y):
     return jnp.clip(x, 0) + jnp.log1p(jnp.exp(-jnp.abs(x))) - x * y
 
 
+def _reshape(x, shape):
+    if isinstance(x, (int, float, np.ndarray, np.generic)):
+        return np.reshape(x, shape)
+    else:
+        return jnp.reshape(x, shape)
+
+
 def promote_shapes(*args, shape=()):
     # adapted from lax.lax_numpy
     if len(args) < 2 and not shape:
@@ -239,7 +249,7 @@ def promote_shapes(*args, shape=()):
     else:
         shapes = [jnp.shape(arg) for arg in args]
         num_dims = len(lax.broadcast_shapes(shape, *shapes))
-        return [lax.reshape(arg, (1,) * (num_dims - len(s)) + s)
+        return [_reshape(arg, (1,) * (num_dims - len(s)) + s)
                 if len(s) < num_dims else arg for arg, s in zip(args, shapes)]
 
 
@@ -252,7 +262,7 @@ def sum_rightmost(x, dim):
     Sum out ``dim`` many rightmost dimensions of a given tensor.
     """
     out_dim = jnp.ndim(x) - dim
-    x = jnp.reshape(x[..., jnp.newaxis], jnp.shape(x)[:out_dim] + (-1,))
+    x = jnp.reshape(jnp.expand_dims(x, -1), jnp.shape(x)[:out_dim] + (-1,))
     return jnp.sum(x, axis=-1)
 
 
@@ -320,8 +330,7 @@ def signed_stick_breaking_tril(t):
     # we omit the step of computing s = z * z_cumprod by using the fact:
     #     y = sign(r) * s = sign(r) * sqrt(z * z_cumprod) = r * sqrt(z_cumprod)
     z = r ** 2
-    z1m_cumprod = jnp.cumprod(1 - z, axis=-1)
-    z1m_cumprod_sqrt = jnp.sqrt(z1m_cumprod)
+    z1m_cumprod_sqrt = jnp.cumprod(jnp.sqrt(1 - z), axis=-1)
 
     pad_width = [(0, 0)] * z.ndim
     pad_width[-1] = (1, 0)
@@ -346,6 +355,17 @@ def clamp_probs(probs):
     return jnp.clip(probs, a_min=finfo.tiny, a_max=1. - finfo.eps)
 
 
+def is_identically_zero(x):
+    """
+    Check if argument is exactly the number zero. True for the number zero;
+    false for other numbers; false for ndarrays.
+    """
+    if isinstance(x, (int, float)):
+        return x == 0
+    else:
+        return False
+
+
 def is_identically_one(x):
     """
     Check if argument is exactly the number one. True for the number one;
@@ -409,7 +429,7 @@ def _von_mises_centered(key, concentration, shape, dtype):
         w = jnp.where(done, w, (1. + s * z) / (s + z))  # Update where not done
 
         y = concentration * (s - w)
-        v = random.uniform(key=uni_vkey, shape=shape, dtype=concentration.dtype, minval=-1., maxval=1.)
+        v = random.uniform(key=uni_vkey, shape=shape, dtype=concentration.dtype)
 
         accept = (y * (2. - y) >= v) | (jnp.log(y / v) + 1. >= y)
 
@@ -428,6 +448,45 @@ def _von_mises_centered(key, concentration, shape, dtype):
     return jnp.sign(u) * jnp.arccos(w)
 
 
+def scale_and_mask(x, scale=None, mask=None):
+    """
+    Scale and mask a tensor, broadcasting and avoiding unnecessary ops.
+    """
+    if is_identically_zero(x):
+        return x
+    if not (scale is None or is_identically_one(scale)):
+        x = x * scale
+    if mask is None:
+        return x
+    else:
+        return jnp.where(mask, x, 0.)
+
+
+# TODO: use funsor implementation
+def periodic_repeat(x, size, dim):
+    """
+    Repeat a ``period``-sized array up to given ``size``.
+    """
+    assert isinstance(size, int) and size >= 0
+    assert isinstance(dim, int)
+    if dim >= 0:
+        dim -= jnp.ndim(x)
+
+    period = jnp.shape(x)[dim]
+    repeats = (size + period - 1) // period
+    result = jnp.repeat(x, repeats, axis=dim)
+    result = result[(Ellipsis, slice(None, size)) + (slice(None),) * (-1 - dim)]
+    return result
+
+
+# src: https://github.com/google/jax/blob/5a41779fbe12ba7213cd3aa1169d3b0ffb02a094/jax/_src/random.py#L95
+def is_prng_key(key):
+    try:
+        return key.shape == (2,) and key.dtype == np.uint32
+    except AttributeError:
+        return False
+
+
 # The is sourced from: torch.distributions.util.py
 #
 # Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
