@@ -9,6 +9,7 @@ import pickle
 import struct
 from urllib.parse import urlparse
 from urllib.request import urlretrieve
+import zipfile
 
 import numpy as np
 
@@ -36,6 +37,11 @@ COVTYPE = dset('covtype', [
 ])
 
 
+DIPPER_VOLE = dset('dipper_vole', [
+    'https://github.com/pyro-ppl/datasets/blob/master/dipper_vole.zip?raw=true',
+])
+
+
 MNIST = dset('mnist', [
     'https://d2hg8soec8ck9v.cloudfront.net/datasets/mnist/train-images-idx3-ubyte.gz',
     'https://d2hg8soec8ck9v.cloudfront.net/datasets/mnist/train-labels-idx1-ubyte.gz',
@@ -105,6 +111,23 @@ def _load_covtype():
     }
 
 
+def _load_dipper_vole():
+    _download(DIPPER_VOLE)
+
+    file_path = os.path.join(DATA_DIR, 'dipper_vole.zip')
+    data = {}
+    with zipfile.ZipFile(file_path) as zipper:
+        data['dipper'] = (
+            np.genfromtxt(zipper.open('dipper_capture_history.csv'), delimiter=',')[:, 1:].astype(int),
+            np.genfromtxt(zipper.open('dipper_sex.csv'), delimiter=',')[:, 1].astype(int)
+        )
+        data['vole'] = (
+            np.genfromtxt(zipper.open('meadow_voles_capture_history.csv'), delimiter=',')[:, 1:],
+        )
+
+    return data
+
+
 def _load_mnist():
     _download(MNIST)
 
@@ -222,6 +245,8 @@ def _load(dset):
         return _load_baseball()
     elif dset == COVTYPE:
         return _load_covtype()
+    elif dset == DIPPER_VOLE:
+        return _load_dipper_vole()
     elif dset == MNIST:
         return _load_mnist()
     elif dset == SP500:
