@@ -79,16 +79,21 @@ results for all the data points, but does so by using JAX's auto-vectorize trans
 from collections import OrderedDict
 import warnings
 
+import numpy as np
+
 from jax import lax, random
 import jax.numpy as jnp
 
 import numpyro
-from numpyro.primitives import Messenger
+from numpyro.distributions.distribution import COERCIONS
+from numpyro.primitives import _PYRO_STACK, Messenger, apply_stack, plate
 from numpyro.util import not_jax_tracer
 
 __all__ = [
     'block',
+    'collapse',
     'condition',
+    'infer_config',
     'lift',
     'mask',
     'reparam',
@@ -98,6 +103,7 @@ __all__ = [
     'seed',
     'substitute',
     'trace',
+    'do'
 ]
 
 
@@ -191,7 +197,7 @@ class replay(Messenger):
         super(replay, self).__init__(fn)
 
     def process_message(self, msg):
-        if msg['name'] in self.guide_trace and msg['type'] in ('sample', 'plate'):
+        if msg['type'] in ('sample', 'plate') and msg['name'] in self.guide_trace:
             msg['value'] = self.guide_trace[msg['name']]['value']
 
 
@@ -201,9 +207,10 @@ class block(Messenger):
     primitive sites  where `hide_fn` returns True from other effect handlers
     on the stack.
 
-    :param fn: Python callable with NumPyro primitives.
-    :param hide_fn: function which when given a dictionary containing
+    :param callable fn: Python callable with NumPyro primitives.
+    :param callable hide_fn: function which when given a dictionary containing
         site-level metadata returns whether it should be blocked.
+    :param list hide: list of site names to hide.
 
     **Example:**
 
@@ -241,6 +248,78 @@ class block(Messenger):
             msg['stop'] = True
 
 
+class collapse(trace):
+    """
+    EXPERIMENTAL Collapses all sites in the context by lazily sampling and
+    attempting to use conjugacy relations. If no conjugacy is known this will
+    fail. Code using the results of sample sites must be written to accept
+    Funsors rather than Tensors. This requires ``funsor`` to be installed.
+    """
+    _coerce = None
+
+    def __init__(self, *args, **kwargs):
+        if collapse._coerce is None:
+            import funsor
+            from funsor.distribution import CoerceDistributionToFunsor
+            funsor.set_backend("jax")
+            collapse._coerce = CoerceDistributionToFunsor("jax")
+        super().__init__(*args, **kwargs)
+
+    def process_message(self, msg):
+        from funsor.terms import Funsor
+
+        if msg["type"] == "sample":
+            if msg["value"] is None:
+                msg["value"] = msg["name"]
+
+            if isinstance(msg["fn"], Funsor) or isinstance(msg["value"], (str, Funsor)):
+                msg["stop"] = True
+
+    def __enter__(self):
+        self.preserved_plates = frozenset(h.name for h in _PYRO_STACK
+                                          if isinstance(h, plate))
+        COERCIONS.append(self._coerce)
+        return super().__enter__()
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        import funsor
+
+        _coerce = COERCIONS.pop()
+        assert _coerce is self._coerce
+        super().__exit__(exc_type, exc_value, traceback)
+
+        if exc_type is not None:
+            return
+
+        # Convert delayed statements to pyro.factor()
+        reduced_vars = []
+        log_prob_terms = []
+        plates = frozenset()
+        for name, site in self.trace.items():
+            if site["type"] != "sample":
+                continue
+            if not site["is_observed"]:
+                reduced_vars.append(name)
+            dim_to_name = {f.dim: f.name for f in site["cond_indep_stack"]}
+            fn = funsor.to_funsor(site["fn"], funsor.Real, dim_to_name)
+            value = site["value"]
+            if not isinstance(value, str):
+                value = funsor.to_funsor(site["value"], fn.inputs["value"], dim_to_name)
+            log_prob_terms.append(fn(value=value))
+            plates |= frozenset(f.name for f in site["cond_indep_stack"])
+        assert log_prob_terms, "nothing to collapse"
+        reduced_plates = plates - self.preserved_plates
+        log_prob = funsor.sum_product.sum_product(
+            funsor.ops.logaddexp,
+            funsor.ops.add,
+            log_prob_terms,
+            eliminate=frozenset(reduced_vars) | reduced_plates,
+            plates=plates,
+        )
+        name = reduced_vars[0]
+        numpyro.factor(name, log_prob.data)
+
+
 class condition(Messenger):
     """
     Conditions unobserved sample sites to values from `data` or `condition_fn`.
@@ -271,11 +350,7 @@ class condition(Messenger):
        >>> assert exec_trace['a']['value'] == -1
        >>> assert exec_trace['a']['is_observed']
     """
-    def __init__(self, fn=None, data=None, condition_fn=None, param_map=None):
-        if param_map is not None:
-            data = param_map
-            warnings.warn("'param_map' argument is renamed to 'data'. We will remove"
-                          " 'param_map' in a future release.", FutureWarning)
+    def __init__(self, fn=None, data=None, condition_fn=None):
         self.condition_fn = condition_fn
         self.data = data
         if sum((x is not None for x in (data, condition_fn))) != 1:
@@ -302,6 +377,24 @@ class condition(Messenger):
             msg['is_observed'] = True
 
 
+class infer_config(Messenger):
+    """
+    Given a callable `fn` that contains Pyro primitive calls
+    and a callable `config_fn` taking a trace site and returning a dictionary,
+    updates the value of the infer kwarg at a sample site to config_fn(site).
+
+    :param fn: a stochastic function (callable containing NumPyro primitive calls)
+    :param config_fn: a callable taking a site and returning an infer dict
+    """
+    def __init__(self, fn=None, config_fn=None):
+        super().__init__(fn)
+        self.config_fn = config_fn
+
+    def process_message(self, msg):
+        if msg["type"] in ("sample",):
+            msg["infer"].update(self.config_fn(msg))
+
+
 class lift(Messenger):
     """
     Given a stochastic function with ``param`` calls and a prior distribution,
@@ -376,11 +469,7 @@ class mask(Messenger):
     :param mask: a boolean or a boolean-valued array for masking elementwise log
         probability of sample sites (`True` includes a site, `False` excludes a site).
     """
-    def __init__(self, fn=None, mask=True, mask_array=None):
-        if mask_array is not None:
-            mask = mask_array
-            warnings.warn("'mask_array' argument is renamed to 'mask'. We will remove"
-                          " 'mask_array' in a future release.", FutureWarning)
+    def __init__(self, fn=None, mask=True):
         if lax.dtype(mask) != 'bool':
             raise ValueError("`mask` should be a bool array.")
         self.mask = mask
@@ -456,12 +545,14 @@ class scale(Messenger):
     This is typically used for data subsampling or for stratified sampling of data
     (e.g. in fraud detection where negatives vastly outnumber positives).
 
-    :param float scale: a positive scaling factor
+    :param scale: a positive scaling factor that is broadcastable to the shape
+        of log probability.
+    :type scale: float or numpy.ndarray
     """
     def __init__(self, fn=None, scale=1.):
         if not_jax_tracer(scale):
-            if scale <= 0:
-                raise ValueError("'scale' argument should be a positive number.")
+            if np.any(np.less_equal(scale, 0)):
+                raise ValueError("'scale' argument should be positive.")
         self.scale = scale
         super().__init__(fn)
 
@@ -474,7 +565,7 @@ class scale(Messenger):
 
 class scope(Messenger):
     """
-    This handler prepend a prefix followed by a ``/`` to the name of sample sites.
+    This handler prepend a prefix followed by a divider to the name of sample sites.
 
     Example::
 
@@ -486,21 +577,23 @@ class scope(Messenger):
        >>>
        >>> def model():
        ...     with scope(prefix="a"):
-       ...         with scope(prefix="b"):
+       ...         with scope(prefix="b", divider="."):
        ...             return numpyro.sample("x", dist.Bernoulli(0.5))
        ...
-       >>> assert "a/b/x" in trace(seed(model, 0)).get_trace()
+       >>> assert "a/b.x" in trace(seed(model, 0)).get_trace()
 
     :param fn: Python callable with NumPyro primitives.
     :param str prefix: a string to prepend to sample names
+    :param str divider: a string to join the prefix and sample name; default to `'/'`
     """
-    def __init__(self, fn=None, prefix=''):
+    def __init__(self, fn=None, prefix='', divider='/'):
         self.prefix = prefix
+        self.divider = divider
         super().__init__(fn)
 
     def process_message(self, msg):
         if msg.get('name'):
-            msg['name'] = f"{self.prefix}/{msg['name']}"
+            msg['name'] = f"{self.prefix}{self.divider}{msg['name']}"
 
 
 class seed(Messenger):
@@ -544,10 +637,7 @@ class seed(Messenger):
        >>> y = handlers.seed(model, rng_seed=1)()
        >>> assert x == y
     """
-    def __init__(self, fn=None, rng_seed=None, rng=None):
-        if rng is not None:
-            warnings.warn('`rng` argument is deprecated and renamed to `rng_seed` instead.', FutureWarning)
-            rng_seed = rng
+    def __init__(self, fn=None, rng_seed=None):
         if isinstance(rng_seed, int) or (isinstance(rng_seed, jnp.ndarray) and not jnp.shape(rng_seed)):
             rng_seed = random.PRNGKey(rng_seed)
         if not (isinstance(rng_seed, jnp.ndarray) and rng_seed.dtype == jnp.uint32 and rng_seed.shape == (2,)):
@@ -557,7 +647,10 @@ class seed(Messenger):
 
     def process_message(self, msg):
         if (msg['type'] == 'sample' and not msg['is_observed'] and
-                msg['kwargs']['rng_key'] is None) or msg['type'] == 'control_flow':
+                msg['kwargs']['rng_key'] is None) or msg['type'] in ['prng_key', 'plate', 'control_flow']:
+            # no need to create a new key when value is available
+            if msg['value'] is not None:
+                return
             self.rng_key, rng_key_sample = random.split(self.rng_key)
             msg['kwargs']['rng_key'] = rng_key_sample
 
@@ -597,11 +690,7 @@ class substitute(Messenger):
        >>> exec_trace = trace(substitute(model, {'a': -1})).get_trace()
        >>> assert exec_trace['a']['value'] == -1
     """
-    def __init__(self, fn=None, data=None, substitute_fn=None, param_map=None):
-        if param_map is not None:
-            data = param_map
-            warnings.warn("'param_map' argument is renamed to 'data'. We will remove"
-                          " 'param_map' in a future release.", FutureWarning)
+    def __init__(self, fn=None, data=None, substitute_fn=None):
         self.substitute_fn = substitute_fn
         self.data = data
         if sum((x is not None for x in (data, substitute_fn))) != 1:
@@ -610,7 +699,7 @@ class substitute(Messenger):
         super(substitute, self).__init__(fn)
 
     def process_message(self, msg):
-        if (msg['type'] not in ('sample', 'param')) or msg.get('_control_flow_done', False):
+        if (msg['type'] not in ('sample', 'param', 'plate')) or msg.get('_control_flow_done', False):
             if msg['type'] == 'control_flow':
                 if self.data is not None:
                     msg['kwargs']['substitute_stack'].append(('substitute', self.data))
@@ -625,3 +714,74 @@ class substitute(Messenger):
 
         if value is not None:
             msg['value'] = value
+
+
+class do(Messenger):
+    """
+    Given a stochastic function with some sample statements and a dictionary
+    of values at names, set the return values of those sites equal to the values
+    as if they were hard-coded to those values and introduce fresh sample sites
+    with the same names whose values do not propagate.
+
+    Composes freely with :func:`~numpyro.handlers.condition` to represent
+    counterfactual distributions over potential outcomes. See Single World
+    Intervention Graphs [1] for additional details and theory.
+
+    This is equivalent to replacing `z = numpyro.sample("z", ...)` with `z = 1.`
+    and introducing a fresh sample site `numpyro.sample("z", ...)` whose value is
+    not used elsewhere.
+
+    **References:**
+
+    1. *Single World Intervention Graphs: A Primer*,
+       Thomas Richardson, James Robins
+
+    :param fn: a stochastic function (callable containing Pyro primitive calls)
+    :param data: a ``dict`` mapping sample site names to interventions
+
+    **Example:**
+
+    .. doctest::
+
+      >>> import jax.numpy as jnp
+      >>> import numpyro
+      >>> from numpyro.handlers import do, trace, seed
+      >>> import numpyro.distributions as dist
+      >>> def model(x):
+      ...     s = numpyro.sample("s", dist.LogNormal())
+      ...     z = numpyro.sample("z", dist.Normal(x, s))
+      ...     return z ** 2
+      >>> intervened_model = handlers.do(model, data={"z": 1.})
+      >>> with trace() as exec_trace:
+      ...     z_square = seed(intervened_model, 0)(1)
+      >>> assert exec_trace['z']['value'] != 1.
+      >>> assert not exec_trace['z']['is_observed']
+      >>> assert not exec_trace['z'].get('stop', None)
+      >>> assert z_square == 1
+    """
+    def __init__(self, fn=None, data=None):
+        self.data = data
+        self._intervener_id = str(id(self))
+        super(do, self).__init__(fn)
+
+    def process_message(self, msg):
+        if msg['type'] != 'sample':
+            return
+        if msg.get('_intervener_id', None) != self._intervener_id and \
+                self.data.get(msg['name']) is not None:
+            if msg.get('_intervener_id', None) is not None:
+                warnings.warn(
+                    "Attempting to intervene on variable {} multiple times,"
+                    "this is almost certainly incorrect behavior".format(msg['name']),
+                    RuntimeWarning)
+            msg['_intervener_id'] = self._intervener_id
+
+            # split node, avoid reapplying self recursively to new node
+            new_msg = msg.copy()
+            apply_stack(new_msg)
+
+            intervention = self.data.get(msg['name'])
+            msg['name'] = msg['name'] + "__CF"  # mangle old name
+            msg['value'] = intervention
+            msg['is_observed'] = True
+            msg['stop'] = True
