@@ -1,8 +1,9 @@
 # Copyright Contributors to the Pyro project.
 # SPDX-License-Identifier: Apache-2.0
 
-from numpyro.infer.elbo import ELBO, RenyiELBO
+from numpyro.infer.elbo import ELBO, RenyiELBO, Trace_ELBO, TraceMeanField_ELBO
 from numpyro.infer.hmc import HMC, NUTS
+from numpyro.infer.hmc_gibbs import HMCECS, DiscreteHMCGibbs, HMCGibbs
 from numpyro.infer.initialization import (
     init_to_feasible,
     init_to_median,
@@ -22,12 +23,17 @@ __all__ = [
     'init_to_uniform',
     'init_to_value',
     'log_likelihood',
+    'DiscreteHMCGibbs',
     'ELBO',
-    'RenyiELBO',
     'HMC',
+    'HMCECS',
+    'HMCGibbs',
     'MCMC',
     'NUTS',
     'Predictive',
+    'RenyiELBO',
     'SA',
     'SVI',
+    'Trace_ELBO',
+    'TraceMeanField_ELBO',
 ]
