@@ -7,13 +7,12 @@ import warnings
 
 import numpy as np
 
-from jax import device_get, lax, random, value_and_grad
+from jax import device_get, jacfwd, lax, random, value_and_grad
 from jax.flatten_util import ravel_pytree
 import jax.numpy as jnp
 
 import numpyro
-import numpyro.distributions as dist
-from numpyro.distributions.constraints import _GreaterThan, _Interval, real, real_vector
+from numpyro.distributions import constraints
 from numpyro.distributions.transforms import biject_to
 from numpyro.distributions.util import is_identically_one, sum_rightmost
 from numpyro.handlers import seed, substitute, trace
@@ -50,7 +49,7 @@ def log_density(model, model_args, model_kwargs, params):
     model_trace = trace(model).get_trace(*model_args, **model_kwargs)
     log_joint = jnp.array(0.)
     for site in model_trace.values():
-        if site['type'] == 'sample' and not isinstance(site['fn'], dist.PRNGIdentity):
+        if site['type'] == 'sample':
             value = site['value']
             intermediates = site['intermediates']
             scale = site['scale']
@@ -119,9 +118,18 @@ def _unconstrain_reparam(params, site):
     if name in params:
         p = params[name]
         support = site['fn'].support
-        if support in [real, real_vector]:
-            return p
         t = biject_to(support)
+        # in scan, we might only want to substitute an item at index i, rather than the whole sequence
+        i = site['infer'].get('_scan_current_index', None)
+        if i is not None:
+            event_dim_shift = t.codomain.event_dim - t.domain.event_dim
+            expected_unconstrained_dim = len(site["fn"].shape()) - event_dim_shift
+            # check if p has additional time dimension
+            if jnp.ndim(p) > expected_unconstrained_dim:
+                p = p[i]
+
+        if support in [constraints.real, constraints.real_vector]:
+            return p
         value = t(p)
 
         log_det = t.log_abs_det_jacobian(p, value)
@@ -168,7 +176,8 @@ def find_valid_initial_params(rng_key, model,
                               enum=False,
                               model_args=(),
                               model_kwargs=None,
-                              prototype_params=None):
+                              prototype_params=None,
+                              forward_mode_differentiation=False):
     """
     (EXPERIMENTAL INTERFACE) Given a model with Pyro primitives, returns an initial
     valid unconstrained value for all the parameters. This function also returns
@@ -232,7 +241,11 @@ def find_valid_initial_params(rng_key, model,
                     key, subkey = random.split(key)
 
         potential_fn = partial(potential_energy, model, model_args, model_kwargs, enum=enum)
-        pe, z_grad = value_and_grad(potential_fn)(params)
+        if forward_mode_differentiation:
+            pe = potential_fn(params)
+            z_grad = jacfwd(potential_fn)(params)
+        else:
+            pe, z_grad = value_and_grad(potential_fn)(params)
         z_grad_flat = ravel_pytree(z_grad)[0]
         is_valid = jnp.isfinite(pe) & jnp.all(jnp.isfinite(z_grad_flat))
         return i + 1, key, (params, pe, z_grad), is_valid
@@ -279,9 +292,9 @@ def _get_model_transforms(model, model_args=(), model_kwargs=None):
                 inv_transforms[k] = biject_to(support)
                 # XXX: the following code filters out most situations with dynamic supports
                 args = ()
-                if isinstance(support, _GreaterThan):
+                if isinstance(support, constraints._GreaterThan):
                     args = ('lower_bound',)
-                elif isinstance(support, _Interval):
+                elif isinstance(support, constraints._Interval):
                     args = ('lower_bound', 'upper_bound')
                 for arg in args:
                     if not isinstance(getattr(support, arg), (int, float)):
@@ -357,11 +370,27 @@ def _guess_max_plate_nesting(model_trace):
     return max_plate_nesting
 
 
+# TODO: follow pyro.util.check_site_shape logics for more complete validation
+def _validate_model(model_trace):
+    # XXX: this validates plate statements under `enum`
+    sites = [site for site in model_trace.values()
+             if site["type"] == "sample"]
+
+    for site in sites:
+        batch_dims = len(site["fn"].batch_shape)
+        if site.get('_control_flow_done', False):
+            batch_dims = batch_dims - 1  # remove time dimension under scan
+        plate_dims = -min([0] + [frame.dim for frame in site["cond_indep_stack"]])
+        assert plate_dims >= batch_dims, \
+            "Missing plate statement for batch dimensions at site {}".format(site["name"])
+
+
 def initialize_model(rng_key, model,
                      init_strategy=init_to_uniform,
                      dynamic_args=False,
                      model_args=(),
-                     model_kwargs=None):
+                     model_kwargs=None,
+                     forward_mode_differentiation=False):
     """
     (EXPERIMENTAL INTERFACE) Helper function that calls :func:`~numpyro.infer.util.get_potential_fn`
     and :func:`~numpyro.infer.util.find_valid_initial_params` under the hood
@@ -379,6 +408,13 @@ def initialize_model(rng_key, model,
         `potential_fn` and `constraints_fn` callables, respectively.
     :param tuple model_args: args provided to the model.
     :param dict model_kwargs: kwargs provided to the model.
+    :param bool forward_mode_differentiation: whether to use forward-mode differentiation
+        or reverse-mode differentiation. By default, we use reverse mode but the forward
+        mode can be useful in some cases to improve the performance. In addition, some
+        control flow utility on JAX such as `jax.lax.while_loop` or `jax.lax.fori_loop`
+        only supports forward-mode differentiation. See
+        `JAX's The Autodiff Cookbook <https://jax.readthedocs.io/en/latest/notebooks/autodiff_cookbook.html>`_
+        for more information.
     :return: a namedtupe `ModelInfo` which contains the fields
         (`param_info`, `potential_fn`, `postprocess_fn`, `model_trace`), where
         `param_info` is a namedtuple `ParamInfo` containing values from the prior
@@ -393,6 +429,10 @@ def initialize_model(rng_key, model,
                                    substitute_fn=init_strategy)
     inv_transforms, replay_model, has_enumerate_support, model_trace = _get_model_transforms(
         substituted_model, model_args, model_kwargs)
+    # substitute param sites from model_trace to model so
+    # we don't need to generate again parameters of `numpyro.module`
+    model = substitute(model, data={k: site["value"] for k, site in model_trace.items()
+                                    if site["type"] in ["param", "plate"]})
     constrained_values = {k: v['value'] for k, v in model_trace.items()
                           if v['type'] == 'sample' and not v['is_observed']
                           and not v['fn'].is_discrete}
@@ -402,6 +442,7 @@ def initialize_model(rng_key, model,
 
         if not isinstance(model, enum):
             max_plate_nesting = _guess_max_plate_nesting(model_trace)
+            _validate_model(model_trace)
             model = enum(config_enumerate(model), -max_plate_nesting - 1)
 
     potential_fn, postprocess_fn = get_potential_fn(model,
@@ -418,15 +459,32 @@ def initialize_model(rng_key, model,
         unconstrained_values = transform_fn(inv_transforms, init_values, invert=True)
         init_strategy = _init_to_unconstrained_value(values=unconstrained_values)
     prototype_params = transform_fn(inv_transforms, constrained_values, invert=True)
-    (init_params, pe, grad), is_valid = find_valid_initial_params(rng_key, model,
-                                                                  init_strategy=init_strategy,
-                                                                  enum=has_enumerate_support,
-                                                                  model_args=model_args,
-                                                                  model_kwargs=model_kwargs,
-                                                                  prototype_params=prototype_params)
+    (init_params, pe, grad), is_valid = find_valid_initial_params(
+        rng_key, model,
+        init_strategy=init_strategy,
+        enum=has_enumerate_support,
+        model_args=model_args,
+        model_kwargs=model_kwargs,
+        prototype_params=prototype_params,
+        forward_mode_differentiation=forward_mode_differentiation)
 
     if not_jax_tracer(is_valid):
         if device_get(~jnp.all(is_valid)):
+            with numpyro.validation_enabled(), trace() as tr:
+                # validate parameters
+                substituted_model(*model_args, **model_kwargs)
+                # validate values
+                for site in tr.values():
+                    if site['type'] == 'sample':
+                        with warnings.catch_warnings(record=True) as ws:
+                            site['fn']._validate_sample(site['value'])
+                        if len(ws) > 0:
+                            for w in ws:
+                                # at site information to the warning message
+                                w.message.args = ("Site {}: {}".format(site["name"], w.message.args[0]),) \
+                                    + w.message.args[1:]
+                                warnings.showwarning(w.message, w.category, w.filename, w.lineno,
+                                                     file=w.file, line=w.line)
             raise RuntimeError("Cannot find valid initial parameters. Please check your model again.")
     return ModelInfo(ParamInfo(init_params, pe, grad), potential_fn, postprocess_fn, model_trace)
 
@@ -554,11 +612,6 @@ class Predictive(object):
                            return_sites=self.return_sites, parallel=self.parallel,
                            model_args=args, model_kwargs=kwargs)
 
-    def get_samples(self, rng_key, *args, **kwargs):
-        warnings.warn("The method `.get_samples` has been deprecated in favor of `.__call__`.",
-                      FutureWarning)
-        return self.__call__(rng_key, *args, **kwargs)
-
 
 def log_likelihood(model, posterior_samples, *args, parallel=False, batch_ndims=1, **kwargs):
     """
@@ -570,13 +623,13 @@ def log_likelihood(model, posterior_samples, *args, parallel=False, batch_ndims=
     :param args: model arguments.
     :param batch_ndims: the number of batch dimensions in posterior samples. Some usages:
 
-        + set `batch_ndims=0` to get prediction for 1 single sample
+        + set `batch_ndims=0` to get log likelihoods for 1 single sample
 
-        + set `batch_ndims=1` to get prediction for `posterior_samples`
+        + set `batch_ndims=1` to get log likelihoods for `posterior_samples`
           with shapes `(num_samples x ...)`
 
-        + set `batch_ndims=2` to get prediction for `posterior_samples`
-          with shapes `(num_chains x N x ...)`
+        + set `batch_ndims=2` to get log likelihoods for `posterior_samples`
+          with shapes `(num_chains x num_samples x ...)`
 
     :param kwargs: model kwargs.
     :return: dict of log likelihoods at observation sites.
@@ -590,13 +643,13 @@ def log_likelihood(model, posterior_samples, *args, parallel=False, batch_ndims=
 
     prototype_site = batch_shape = None
     for name, sample in posterior_samples.items():
-        if batch_shape is not None and sample.shape[:batch_ndims] != batch_shape:
+        if batch_shape is not None and jnp.shape(sample)[:batch_ndims] != batch_shape:
             raise ValueError(f"Batch shapes at site {name} and {prototype_site} "
                              f"should be the same, but got "
                              f"{sample.shape[:batch_ndims]} and {batch_shape}")
         else:
             prototype_site = name
-            batch_shape = sample.shape[:batch_ndims]
+            batch_shape = jnp.shape(sample)[:batch_ndims]
 
     if batch_shape is None:  # posterior_samples is an empty dict
         batch_shape = (1,) * batch_ndims
